-- Fischer mutual exclusion example
-- error trace generation

define(delay_a,5)		-- max delay time to register intent
define(delay_b,6)		-- min time to delay before rechecking

var 
	x,      	-- P1's clock
	y      		-- P2's clock
	: analog;
	k: discrete;  	-- whose turn it is (has values 0,1,2)

 -- -------------------------------------------------------------- *)

automaton p1
synclabs : start_1, set_k_1, enter_cs_1, reset_1;

initially loc_1 & True;

loc loc_1: while True wait {dx in [4/5,1]}
	when k=0 do {x' =0} sync start_1 goto loc_2;
	
loc loc_2: while x<=delay_a wait {dx in [4/5,1]}
	when True do {k' =1,x' =0} sync set_k_1 goto loc_3;

loc loc_3: while True wait {dx in [4/5,1]}
	when x>=delay_b & k=1 sync enter_cs_1 goto cs;
	-- two failed attempts
	when x>=delay_b & k=0 sync reset_1 goto loc_1;
	when x>=delay_b & k=2 sync reset_1 goto loc_1;

loc cs: while True wait {dx in [4/5,1]}
	when True do {k' = 0} sync reset_1 goto loc_1;
end


automaton p2
synclabs : start_2, set_k_2, enter_cs_2, reset_2;

initially loc_1 & True;

loc loc_1: while True wait {dy in [1,11/10]}
	when k=0 do {y' =0} sync start_2 goto loc_2;
	
loc loc_2: while y<=delay_a wait {dy in [1,11/10]}
	when True do {k' =2, y' =0} sync set_k_2 goto loc_3;

loc loc_3: while True wait {dy in [1,11/10]}
	when y>=delay_b & k=2 sync enter_cs_2 goto cs;
	-- two failed attempts
	when y>=delay_b & k=0 sync reset_2 goto loc_1;
	when y>=delay_b & k=1 sync reset_2 goto loc_1;

loc cs: while True wait {dy in [1,11/10]}
	when True do {k' = 0} sync reset_2 goto loc_1;
end

--

var 
   init_reg, final_reg, reached, reached_viol : region;

init_reg := loc[p1] = loc_1 & loc[p2] = loc_1 & k=0;

final_reg := loc[p1] = cs & loc[p2] = cs;

reached := reach forward from init_reg endreach;

reached_viol := reached & final_reg;

if empty(reached_viol)
  then prints "Mutual exclusion requirement holds";
  else prints "Mutual exclusion violated";
       print trace to final_reg using reached;
endif;





