-- Nuclear reactor 
--
-- Clock translation
--

-- system models 2 cooling rods, and temperature of reactor core
--
-- there is a delay between removing a cooling rod, and reinserting it
-- rod gets removed when temperature is too high.

-- temperature x 
--     rises at rate x/10 + 50 when no rods present
--     decreases at rate x/10 -56 when rod 1 used
--     decreases at rate x/10 -60 when rod 2 used
-- rod cannot be reused until delay of T.
-- low temp = 510?
-- high temp = 550?
--
-- Goal: a rod is always available to keep tempeture between 510 and 550
-- Parametric analysis: find values of T for which above holds.



define(ten_log_5_plus, 161/10)	-- 1.6094
define(ten_log_5_minus, 16)
define(ten_log_9_5_plus, 59/10 )	-- 0.5878
define(ten_log_9_5_minus, 58/10)



var 
   y1,            -- timer for rod 1
   y2             -- timer for rod 2
     : clock;
   x              -- clock-translated variable from temperature
     : clock;

   T              -- minimal time delay before reusing a cooling rod
     : parameter;

-- model delay in time for rod_1 to be used again for cooling
automaton rod_1
synclabs: add_1, remove_1;
initially out_1 & y1 = T;

loc out_1: while True wait {}
    when y1 >= T sync add_1 goto in_1;

loc in_1: while True wait {}
    when True sync remove_1 do {y1' = 0} goto out_1;

end -- rod_1



-- model delay in time for rod_2 to be used again for cooling
automaton rod_2
synclabs: add_2, remove_2;
initially out_2 & y2 = T;


loc out_2: while True wait {}
    when y2 >= T sync add_2 goto in_2;

loc in_2: while True wait {}
    when True sync remove_2 do {y2' = 0} goto out_2;

end -- rod_2



-- model temperature via clock-translated variable x
automaton temp
synclabs: add_1, remove_1, add_2, remove_2;
initially no_rod & x = 0;

loc no_rod: while x <= ten_log_5_plus wait {}
    when ten_log_5_minus <= x & x <= ten_log_5_plus 
        sync add_1 do {x' = 0} goto rod_1;
    when ten_log_5_minus <= x & x <= ten_log_5_plus 
        sync add_2 do {x' = 0} goto rod_2;



loc rod_1: while x <= ten_log_5_plus wait {}
    when ten_log_5_minus <= x & x <= ten_log_5_plus 
        sync remove_1 do {x' = 0} goto no_rod;

loc rod_2: while x <= ten_log_9_5_plus wait {}
    when ten_log_9_5_minus <= x & x <= ten_log_9_5_plus 
        sync remove_2 do {x' = 0} goto no_rod;

end -- temp


-- analysis commands

var 
    init_reg, final_reg, b_reached : region;


init_reg :=   loc[rod_1] = out_1 & y1 = T
            & loc[rod_2] = out_2 & y2 = T
            & loc[temp]  = no_rod & x = 0;


-- check the exact specification P-H used?

final_reg :=   loc[temp] = no_rod & ten_log_5_minus <= x & x <= ten_log_5_plus
             & loc[rod_1] = out_1 & y1 <= T
             & loc[rod_2] = out_2 & y2 <= T;

b_reached := reach backward from final_reg endreach;

prints "Control rod NOT available under the following conditions";
print omit all locations hide non_parameters in b_reached & init_reg endhide;
