//
// C++ code for converting from C internal structure (post format) to 
//      C++ internal structure (post info only).
//


// Notes 3/95. on inf. assignments.
// Fix nr_constraints now takes into account the inf. bds which do not give
//    any constraints.  inf. bds detected by leading entry n[0] = 0.
// also need to fix set_up_assignment's fixing of the matrix.

#include "convert.H"


// put in *mcpp the C++ matrix for m
// assume space allocated for *mcpp , but NOT for its data in p[][]
void
Cpp_matrix(matrix *mcpp, struct _p_matrix *m) {
  int i,j;
  struct _p_vector *vec;      // for the dummy vector passing through list of
                              // vectors in matrices.
  vec = m->vectors;
  *mcpp = matrix(m->rows, m->columns);
  for (i=0;i<m->rows;i++) { 
    for (j=0;j<m->columns;j++) {
      // info stored in numerators only for matrices?? 
      mcpp->p[m->rows - i - 1][j] = vec->n[j];
    }
    vec = vec->next;
  }
  mcpp->nonstrict = !mcpp->is_strict();
}

// this creates the nplus1 form of the matrix in *mcpp_nplus1
void
Cpp_matrix_nplus1(matrix *mcpp_nplus1, struct _p_matrix *m) {
  matrix mcpp;
  Cpp_matrix(&mcpp,m);
  *mcpp_nplus1 =  *my_nplus1_matrix(mcpp);
}

/*
// create the nonstrict nplus1_form in *mcpp_ns_nplus1
void
Cpp_matrix_ns_nplus1(ns_matrix *mcpp_ns_nplus1, struct _p_matrix *m) {
  matrix mcpp;
  Cpp_matrix(&mcpp,m);
  *mcpp_ns_nplus1 =  *ns_nplus1_matrix(mcpp);
}
*/


int
nr_locations_in(struct automata_list C_a) {
  int c;                        //  location count
  struct info_location *loc;    // dummy locn for traversing list of locations
  c = 0;
  loc = C_a.locations;
  while (loc != NULL) {
    c++;
    loc = loc->next;
  }
//  cout << "Number of locations is " << c << "\n";
  return (c);
}

// returns whether a particular vector is a slope vector, as opposed to an 
//    assignment vector
// assumes input is non-null
Bool
is_slope_vector(struct _p_vector *v) {
  return (v->variable == -1);
}

// returns the number of vectors that correspond to slopes in C_a
int
nr_slopes_in(struct automata_list C_a) {
  int c;                        //  slope vector count
  struct _p_vector  *vec;    // dummy vec for traversing list of vectors
  c = 0;
  vec = C_a.vectors;
  while (vec != NULL) {
    if (is_slope_vector(vec) )
      c++;
    vec = vec->next;
  }
//  cout << "Number of slope_vectors is " << c << "\n";
  return (c);
}

// make space for, and copy over slopes from C_a
void
automaton::copy_slopes(struct automata_list C_a) {
// note C vectors have variable id we can ignore in slope vectors
// it also has *d and *n, instead of single number and common denominator?
// seems like we can ignore the *d in slope vectors too, see output.c pw()
// cope with BUG in slopes in C parser.  first entry must be 1, not 0.
// *****
  int i;
  int c = 0;                  // index position in array
  struct _p_vector  *vec;     // dummy vec for traversing list of vectors
  int max_vect_id = -1;

   nr_slopes = nr_slopes_in(C_a);
   slopes = ( vector *) malloc (nr_slopes * sizeof(vector));

  // find the max vect id,. so we know how much space to store in reverse index
  vec = C_a.vectors;
  while (vec != NULL) {
    if (vec->vect_id > max_vect_id) {
      max_vect_id = vec->vect_id;
    }
    vec = vec->next;
  }

//  cout << "max_vect_id = " << max_vect_id << "\n";
  // need one extra for off-by-one.
  id_slope_index = (int *) malloc ((max_vect_id+1) * sizeof(int));
 for (i=0;i<max_vect_id+1;i++) {
   id_slope_index[i] = -1;
 }
 
  vec = C_a.vectors;
  while (vec != NULL) {
    if (is_slope_vector(vec)) {
      slopes[c] = vector(vec->len); 
      for (i=0;i<vec->len;i++) {
	slopes[c].p[i] = vec->n[i];
      }
      // store the lcm in constant column
      slopes[c].p[vec->len-1] = vec->vlcm;
// cout << "assigning slope index for slope_id " << vec->vect_id << " the ptr to slopes as " << c << " \n";
      // **** cope with BUG in C parser, need first entry = 1.
      // ** bug fixed at source, i.e. in parsing phase
      // ** don't set it now, since there are real lines as slopes for
      // ** vars whose slopes are not declared.
      // slopes[c].p[0]=1;

      id_slope_index[vec->vect_id] = c;
      c++;
    }
    else { //  cout << "this vector not a slope " << vec->vect_id << "\n";
	 }
    vec = vec->next;
  }
  cout << flush;
}


void
automaton::copy_sync_labels(struct  automata_list C_a) {
  int i;
  struct var_name *evt_name_p = C_a.event_labels;
// cout << "number of event labels = " << C_a.nr_event_labels << "\n" << flush;
  sync_label_names = (char **) malloc (C_a.nr_event_labels * sizeof (char *));
  nr_sync_labels = C_a.nr_event_labels;
  for (i=0;i<C_a.nr_event_labels;i++) {
//    trace_out("doing label",i);
    sync_label_names[i] = str_preserve(evt_name_p->name);
    evt_name_p = evt_name_p->next;
  }
// trace_out("done with sync labels ",0);
}




int
nr_matrices_in(struct automata_list C_a) {
  int c;                        //  slope vector count
  struct _p_matrix  *mp;    // dummy matrix ptr for traversing list of matrices
  c = 0;
  mp = C_a.matrices;
  while (mp != NULL) {
    c++;
    mp = mp->next;
  }
//  cout << "Number of matrices is " << c << "\n";
  return (c);
}



// make space for, and copy over matrices from C_a
// don't create matrices for assignments quite yet.
// beware: this will now include matrices for the postslopes?
void
automaton::copy_matrices(struct automata_list C_a) {
  int c = 0;                  // index position in array
  struct _p_matrix  *mp;      // dummy matrix ptr for traversing list of mtr.
  int nr_matrices = 0;
  int max_matrix_id = -1;

   nr_matrices = nr_matrices_in(C_a);
  nr_matrices_in_store = nr_matrices;
   matrix_store = ( matrix *) malloc (nr_matrices * sizeof(matrix));

  // find the max matrix id,. to allocate enough space in reverse index
  mp = C_a.matrices;
  while (mp != NULL) {
    if (mp->matrix_number > max_matrix_id) {
      max_matrix_id = mp->matrix_number;
    }
    mp = mp->next;
  }
  id_matrix_index = (int *) malloc ((max_matrix_id+1) * sizeof(int));

  mp = C_a.matrices;
  while (mp != NULL) {
    Cpp_matrix(&(matrix_store[c]),mp);
    id_matrix_index[mp->matrix_number] = c;
    mp = mp->next;
    c++;
  }
}


int
nr_trans_in(struct info_location *locn) {
// return the number of transitions out of this location 
  int count = 0;
  struct info_transition *transp;
  transp = locn->transitions;
  while (transp != NULL) {
    count++;
    transp = transp->next;
  }
  return( count );
}

// return ptr to vector with this ID
//  cannot assume the vector is in the list of slope vectors, could be an
//  assignmt vector.
struct _p_vector *
ptr_to_vector(int vectid, struct _p_vector *vector_list) 
{
  // look on C automaton's list of vectors
  struct _p_vector *pv = vector_list;
  Bool found = false;
  while (!found && (pv != NULL)) {
    if (!(found = (pv->vect_id == vectid))) 
      pv = pv->next;
  }
  if (found) {
    return( pv );
  }
  else {
    cerr << "Internal error: Vector with ID " << vectid << " not found \n" << flush;
    exit(1);	
  }
}


Bool
vector_eq(struct _p_vector *pv1, struct _p_vector *pv2) {
   int nr_cols = pv1->len;
   if (pv2->len != nr_cols) {
     cout << "Inconsistent vectors being compared\n" << flush;
     exit(1);
   }
   else {
     int i;
     Bool eq = true;
     for (i=0;i<nr_cols;i++) {
       if (pv1->n[i] != pv2->n[i]) {
	 eq = false;
       }
     }
     return(eq);
   }
 }

// need to pass in C_a to have access to list of its vectors
int
automaton::nr_constraints_in(struct info_assign *asp, 
			     struct automata_list C_a) 
{
  // test if the vectors are the same.  if yes, then only need one assignment
  // and it will be equals.
  // currently no facility for single upper bound,or lower bound.
  // otherwise, we'll need two, an upper bound, and a lower bound.
  struct _p_vector *pv1, *pv2;
  pv1 = ptr_to_vector(asp->vect1,C_a.vectors);
  pv2 = ptr_to_vector(asp->vect2,C_a.vectors);
//  cout << "first vector is " << asp->vect1 << " and second " 
//    << asp->vect2 << "\n" << flush;
  if (vector_eq(pv1,pv2)) {
    // check there' not both infinite
    if (pv1->n[0] == 0) {
      return(0);
    }
    else {
      return(1);
    }
  }
  else {
    // two vectors are different
    if ((pv1->n[0]==0) || (pv2->n[0] == 0)) {
      // exactly one bound must be infinite, since both not the same
      return(1);
    }
    else {
      return(2);
    }
  }
}

// return 0 if neither bd is infinite, 2 if neg only, 1 if pos only, 3 for both
int
magic_inf_nr(struct info_assign *asp, struct automata_list C_a){
  struct _p_vector  *vecp1 = ptr_to_vector(asp->vect1,C_a.vectors);
  struct _p_vector  *vecp2 = ptr_to_vector(asp->vect2,C_a.vectors); 
  int infnr= 0;
  if (vecp1->n[0]==0) {
    infnr += 2;
  }
  if (vecp2->n[0]==0) {
    infnr++;
  }
  return(infnr);
}

// given a ptr to a transition (in C), set up the info for C++ version
// with multiple assignments.  need to assign a matrix on the free store
// for the assignment, and return the value of the ptr to it in ass_mp
// the calling routine will have to manage these on a list somewhere
// to allow sharing, but don't worry about this for now.
void
automaton::old_set_up_assignment(struct info_transition *trp, 
			     struct automata_list C_a,
		  int *nbvars, int **indp, matrix **ass_mp) 
{
  struct info_assign *asp = trp->assign;
  struct _p_vector *vecp;
  matrix *mp;
  int varcount = 0;       // how many variable are being reassigned
  int constr_count = 0;   // number of constraints needed in reassignments
  int global_num_cols = columns + 1; // changed 9/19/95, HWT since creating
                                     // an n+1 form matrix instead
//  int global_num_cols = columns; // dangerous assumption that # columns
                                 // is stored in the parser's var "columns"
  int ass_nr;            // for counting through assignments  
  int row;               // row of matrix to be added
  int i;                 // generic loop index

  // figure out how many rows to create in ass matrix.
  while (asp != NULL) 
    {
      varcount++;
      constr_count += nr_constraints_in(asp,C_a);
      asp=asp->next;
    }
  *nbvars = varcount;
  *indp = (int *) malloc (varcount * sizeof(int));
  // matrix needs a row for the 1 0 0 0 1 row at the end
#ifdef DEBUG_CONVERT
cout << " constr_count = " << constr_count << "\n";
cout << " global_num_cols varcount = " <<  global_num_cols << ", " << varcount << "\n";
#endif
  mp = new matrix(constr_count+1, global_num_cols +varcount);

  // above constructor does not add the ones at extremes of final row
  // nor fill in the zeros
  for (i=1;i<global_num_cols+varcount-1;i++) {
    mp->p[constr_count][i] = 0;
  }
  mp->p[constr_count][0] = 
	mp->p[constr_count][global_num_cols+varcount-1] = 1;

  // put in correct values for assignment matrix
  asp = trp->assign;
  ass_nr = 0; 
  row = 0;
  while (asp != NULL) 
    {
      int nr_ass_vecs;     // nr of different assignment vecs for this ass.
      // add in the indices.
      (*indp)[ass_nr] = asp->var_id;

      // add in the constraints from the assignment vectors and indices!
      // for the POST assignment, the added variables are the primed ones.
      // recall in constraint form, first elt of row = 0 for equality
      //       first elt of assignmt vector is denom of constants
      nr_ass_vecs = nr_constraints_in(asp,C_a);
      int magic_nr = magic_inf_nr(asp,C_a); 
      vecp = ptr_to_vector(asp->vect1,C_a.vectors);
      if ((nr_ass_vecs == 1) && (magic_nr ==0)) {
	//     dealing with an equality assignment
	  // i.e. none are infinite, we have an equality
	  mp->p[row][0] = 0;   // for equality
	  // don't copy the constant (last value) here
// changed HWT 9/20 for strictness column
	  for (i=1;i<global_num_cols-2;i++) {
	    mp->p[row][i] = vecp->n[i];
	  }
	  // added line here to set this col's entry to 0.
 	  mp->p[row][global_num_cols-2] = 0;
	  // blank out all except the last column for the constant
	  for (i=global_num_cols-1;i<global_num_cols+varcount-1;i++) {
	    mp->p[row][i] = 0;
	  }
	  // the value for the primed assigned var
	  // it's negated since comes from opposite side of = sign.
	  mp->p[row][global_num_cols-1+ass_nr] = -vecp->n[0];
	  // the constant attached to the end of the row
	 // changed HWT, 9/20/95, since global_num_cols is 1 bigger to include
	 //  entry for strictness variable.
	  mp->p[row][global_num_cols+varcount-1] = vecp->n[global_num_cols-2];
	  row++;
      }
      else {
	// not an equality assignment, but strictly an interval.
	// treat each bound independently
	// dealing with an upper and lower bound
	// assume the lower/upper bound are non-strict?
	// 
	// first the lower
	// 
	if ((magic_nr == 1) || (magic_nr ==0)) {
	  // the lower bound is not infinite
	  mp->p[row][0] = 1;   // for inequality
	  // negate all values, since x+y+b <= x' is same as x' -(x+y+b) >=0.
	  // don't copy the constant (last value) here
	  for (i=1;i<global_num_cols-1;i++) {
	    mp->p[row][i] = -vecp->n[i];
	  }
	  // blank out all except the last column for the constant
	  for (i=global_num_cols-1;i<global_num_cols+varcount-1;i++) {
	    mp->p[row][i] = 0;
	  }
	  // the value for the primed assigned var
	  // it's not negated, see above
	  mp->p[row][global_num_cols-1+ass_nr] = vecp->n[0];
	  // the constant attached to the end of the row (negated)
	  mp->p[row][global_num_cols+varcount-1] = -vecp->n[global_num_cols-1];
	  row++;
	}

	// then the upper
	// move to second vector
	vecp = ptr_to_vector(asp->vect2,C_a.vectors);
	// 
	if ((magic_nr==2) || (magic_nr==0)) {
	  // then the upper bound is not infinite
	  mp->p[row][0] = 1;   // for inequality
	  // copy all values, since x+y+b >= x' is same as(x+y+b)-x' >=0.
	  // don't copy the constant (last value) here
	  for (i=1;i<global_num_cols-1;i++) {
	    mp->p[row][i] = vecp->n[i];
	  }
	  // blank out all except the last column for the constant
	  for (i=global_num_cols-1;i<global_num_cols+varcount-1;i++) {
	    mp->p[row][i] = 0;
	  }
	  // the value for the primed assigned var
	  // it's negated, see above
	  mp->p[row][global_num_cols-1+ass_nr] = -vecp->n[0];
	  // the constant attached to the end of the row
	  mp->p[row][global_num_cols+varcount-1] = vecp->n[global_num_cols-1];
	  row++;
	}
      }
      ass_nr++;
      asp = asp->next;
    }
  mp->nonstrict = !(mp->nplus1_is_strict(global_num_cols-1));
  *ass_mp = mp;

}



// given a ptr to a transition (in C), set up the info for C++ version
// with multiple assignments.  need to assign a matrix on the free store
// for the assignment, and return the value of the ptr to it in ass_mp
// the calling routine will have to manage these on a list somewhere
// to allow sharing, but don't worry about this for now.
// first 2 param are input
// last 3 are output
void
automaton::set_up_assignment(struct info_transition *trp, 
			     struct automata_list C_a,
		  int *nbvars, int **indp, matrix **ass_mp) 
{
  struct info_assign *asp = trp->assign;
  struct _p_vector *vecp;
  matrix *mp;
  int nr_ass = 0;       // how many variable are being reassigned
  int constr_count = 0;   // number of constraints needed in reassignments
  int global_num_cols = columns + 1; // changed 9/19/95, HWT since creating
                                     // an n+1 form matrix instead
//  int global_num_cols = columns; // dangerous assumption that # columns
                                 // is stored in the parser's var "columns"
  int ass_nr;            // for counting through assignments  
  int row;               // row of matrix to be added
  int i;                 // generic loop index
  int *reassigned = (int *) malloc (sizeof(int) * var_count);
  clear_int_array(reassigned,var_count);
  // figure out how many rows and cols to create in ass matrix.
/*
  // this looks at constraint matrix, but we need to use parser info instead
  while (asp != NULL) 
    {
      struct _p_vector *vec = ptr_to_vector(asp->vect1, C_a.vectors);
      for (i=1;i<=var_count;i++) {
        if (vec->n[var_count+i] != 0) {
	  reassigned[i-1] = 1;
	}
      }
      constr_count++;
      asp=asp->next;
    }
*/
  // trp->ass_vars may be NULL if there are no reassignments
  if (trp->assign != NULL) {
    for (i=0;i<var_count;i++) {
      reassigned[i] = trp->ass_vars[i];
    }
  }

  for (i=0;i<var_count;i++) {
    if (reassigned[i]) 
      nr_ass++;
  }

  *nbvars = nr_ass;
  *indp = (int *) malloc (nr_ass * sizeof(int));
  int ass_ind = 0;
  for (i=0;i<var_count;i++) {
    if (reassigned[i]) {
      // use 1 for 0'th var, etc.
      (*indp)[ass_ind] = i+1;
      ass_ind++;
    }
  }
  while (asp != NULL) 
    {
      constr_count++;
      asp=asp->next;
    }
  


  // matrix needs a row for the 1 0 0 0 1 row at the end
  mp = new matrix(constr_count+1, global_num_cols +nr_ass);

  // above constructor does not add the ones at extremes of final row
  // nor fill in the zeros
  mp->set_last_row();

  // put in correct values for assignment matrix
  asp = trp->assign;
  ass_nr = 0; 
  row = 0;
  while (asp != NULL) 
    {
      // add in the constraints from the assignment vectors and indices!
      // for the POST assignment, the added variables are the primed ones.
      // recall in constraint form, first elt of row = 0 for equality
      // directly making an n+1 form

      vecp = ptr_to_vector(asp->vect1,C_a.vectors);
      if (vecp->n[0] == 0) {
	mp->p[row][0] = 0;
      }
      else {
	mp->p[row][0]=1;
      }
      // copy everything for the var_count real vars 
      for (i=1;i<=var_count;i++) {
	mp->p[row][i] = vecp->n[i];
      }
      // fix the column for the eps_slack variable
      if (vecp->n[0]==2)  {
	mp->p[row][var_count+1] = -1;
      }
      else {
	mp->p[row][var_count+1] = 0;
      }
      // copy rest of assigned vars 
      for (i=0;i<nr_ass;i++) {
#ifdef DEBUG_CONVERT
cout << "*** copying from index " << var_count+(*indp)[i] << " to ind " <<
var_count+2+i << " with value " << vecp->n[var_count+(*indp)[i]] << "\n" << flush;
#endif
	mp->p[row][var_count+2+i] = vecp->n[var_count+(*indp)[i]];
      }
      // copy constant
      mp->p[row][global_num_cols+nr_ass-1] = vecp->n[2*var_count+1];
      row++;
      asp = asp->next;
    }
  mp->nonstrict = !(mp->nplus1_is_strict(global_num_cols-2));
  *ass_mp = mp;

}



location *
automaton::dest_location_ptr(char *dest_name) 
// don't use Greg's loc_index lookup in location.c.
// his assumes the list is called locations.

// here: assumes the location names have already been installed
{
  int i=0;
  Bool found=false;
  
  while ((i<nr_locations) && !found) {
    // strcmp return 0 if both strings the same
    if (! (found = !strcmp(dest_name,locations[i].name) ) ) 
      i++;
  }
  if (!found) {
    cout << "location: " << dest_name << " not found \n";
    exit(2);
  }
  else {
//    cout << "location: " << dest_name << " found at index " << i << "\n" << flush;
      return( &locations[i] );
    }
}



void
automaton::copy_locations(struct automata_list C_a) {
// set the post-vectors, ptr to matrix for invariant 
// allocate space for transitions, and copy these, creating matrices for asst.
// ignore sync labels for now.
// assumes the nr_locations has been correctly assigned already.
//    and the space allocated for them.
// only need to set up post- info.
  int i,j;
  struct info_location *loc;
  struct info_transition *tr;

  loc = C_a.locations;
  for (i=0;i<nr_locations;i++) {
    // copy over location names first, for access when setting up transitions
    // should really use the correct constructors
    locations[i].name = str_preserve(loc->location_name);
    locations[i].locn_id = loc->location_number;
    locations[i].asap = false;
    loc = loc->next;
  }
  loc = C_a.locations;
  for (i=0;i<nr_locations;i++) {
    // copy over each location

    // cout << "location number = " << loc->location_number << "\n";

    // slopes
    // using slope matrices now instead of lists of slopes
   

    // set up poly from the constraint matrix for slopes

   matrix *m_sl;
    m_sl = nplus1_rates_matrix(matrix_store[id_matrix_index[loc->slope_matrix_id]]);
    matrix_store[id_matrix_index[loc->slope_matrix_id]].free();
    matrix_store[id_matrix_index[loc->slope_matrix_id]] = *m_sl;
    delete m_sl;
    // avoid usual constructor, which simplifies for strictness.
    locations[i].post_slopes_matrix = 
	new polyhedron();
    locations[i].post_slopes_matrix->nplus1_form =
      new lib_polyhedron(matrix_store[id_matrix_index[loc->slope_matrix_id]],true);
    locations[i].post_slopes_matrix->strict = 
	matrix_store[id_matrix_index[loc->slope_matrix_id]].is_strict();
    locations[i].post_slopes_matrix->eps_pos = var_count+1;


    locations[i].pre_slopes_matrix = NULL;
/*
    locations[i].nr_slopes = loc->slop_end_vect - loc->slop_start_vect + 1;
// cout << "number slopes for location " << i << " is " << 
// locations[i].nr_slopes << "\n";
    locations[i].post_slopes = (slope_vector **) malloc 
      ( locations[i].nr_slopes * sizeof( slope_vector *) );
    // copy ptrs to slope vectors
    for (j=0;j<locations[i].nr_slopes;j++) {
// cout << "slope " << j << " is in position " << loc->slop_start_vect+j << 
//  " with id index " << id_slope_index[loc->slop_start_vect+j] << "\n";
      locations[i].post_slopes[j] = 
	&slopes[id_slope_index[loc->slop_start_vect+j]];
    }
// cout << "\n";
*/
    // matrix invar.
     matrix *m_inv;

    m_inv = 
	my_nplus1_matrix(matrix_store[id_matrix_index[loc->matrix_inv_id]]);
// redundant    m_inv.nonstrict = ! m_inv->nplus1_is_strict();

    matrix_store[id_matrix_index[loc->matrix_inv_id]].free();
    matrix_store[id_matrix_index[loc->matrix_inv_id]] = *m_inv;
    delete m_inv;
    locations[i].invariant = 
	&(matrix_store[id_matrix_index[loc->matrix_inv_id]]);

    // trans
    locations[i].nr_post_trans = nr_trans_in(loc);
    locations[i].nr_pre_trans = 0;
//    cout << "number transitions = " << locations[i].nr_post_trans << 
//      "\n" << flush;
    locations[i].post_trans = 
      (transition *) malloc ( locations[i].nr_post_trans *
					sizeof(transition));
    locations[i].pre_trans = NULL;
    tr = loc->transitions;    
    for (j=0;j<locations[i].nr_post_trans;j++) {
      // set up the assignment matrix first:as nbvars, ind, ass_mp
      int nbvars;  // number vars being assigned
      int *ind;    // vector of indices of the assigned vars
      matrix *ass_mp; // ptr to matrix for the assignment.  must be put on list
                      // called matrix_hd.
      // set up the above params
      set_up_assignment(tr,C_a,&nbvars,&ind,&ass_mp);

     matrix *m_gd;


    
    m_gd = nplus1_plain_matrix(matrix_store[id_matrix_index[tr->matrix_guard_id]],
	matrix_store[id_matrix_index[tr->matrix_guard_id]].nbcolumns-1);
    
    matrix_store[id_matrix_index[tr->matrix_guard_id]].free();
    matrix_store[id_matrix_index[tr->matrix_guard_id]] = *m_gd;
    delete m_gd;
      
      locations[i].post_trans[j] = 
	transition(&locations[i], dest_location_ptr(tr->destination), 
		   tr->synlabel,
		   &matrix_store[id_matrix_index[tr->matrix_guard_id]],
		   tr->was_asap,
		   new assignment(nbvars, ind, ass_mp));
      if (tr->was_asap) {
	// assume that any ASAP transition is legitimate syntax.
	// i.e. has no guard, and no conflicting assignments, etc.
	locations[i].asap = true;
      }
      tr = tr->next;
    }
    loc =loc->next;
  }

}


// handle initial information.
void
automaton::copy_init_info(struct automata_list C_a) {
// need to copy over info for the initial location
//    and set up ptr to the inital region.
// assumes the matrix list, and location lists have been copied over already
  // find the matrix in the matrix store.

     matrix *m_init;
    m_init = my_nplus1_matrix(matrix_store[id_matrix_index[C_a.init_m_id]]);
    matrix_store[id_matrix_index[C_a.init_m_id]].free();
    matrix_store[id_matrix_index[C_a.init_m_id]] = *m_init;
    delete m_init;

  init_region_p = &matrix_store[id_matrix_index[C_a.init_m_id]];
 
  // find location by name.
  init_location_p = dest_location_ptr(C_a.init_loc_name);
}


void
automaton::adjust_locn_ids() {
  int i;
  for (i=0;i<nr_locations;i++) {
    locations[i].locn_id = i;
  }
}

// also defined in automaton.C!
#define MAX_VARS 150

// use info in var_head to make the list of var_names in varnames[]
void
make_var_names() {
  if (var_count > MAX_VARS) {
    cout << "Error. too many vars\n";
    cout << "Exceeds limit of " << MAX_VARS << endl;
    exit(1);
  }
  else {
/*

    varnames = (char **) malloc (var_count * sizeof(char *));
*/
    struct var_name *vn = var_head;
    while (vn != NULL) {
      // indices are in reverse order, start at 1. want to rename to start at 0
      varnames[vn->index-1] = str_preserve(vn->name);
//      cout << "putting " << vn->name << " into pos " << vn->index-1 << "\n";
      vn = vn->next;
    }
  }
}


// constructor that creates a C++ automaton out of a C_automaton
automaton::automaton(struct automata_list C_automaton) {
// assume the C_automaton list is valid input.
   int nr_locs;

   // use var_head to make the varnames array
//    make_var_names();

   // set up space for the locations
   nr_locs = nr_locations_in(C_automaton);
   *this = automaton(C_automaton.automaton_name,nr_locs);
   nr_locations = nr_locs;

   copy_sync_labels(C_automaton);

   // do the global list of slopes
   // was copy_slopes(C_automaton);
   // no longer necessary, since the slopes are now matrices


   // do the global list of matrices
   copy_matrices(C_automaton);

   copy_locations(C_automaton);

   copy_init_info(C_automaton);


   // now make the index numbers of locations match the ID numbers.
   // since the actual location numbers are never used within the structure
   //  this is safe.
   adjust_locn_ids();
   // set asap field for all locations.
   set_location_asap();
   
  // free id index vectors
  free_id_vectors();

}


// create a list of C++ automata from a list of C_automata
// always some data in the first automaton in the "list"
automaton_info::automaton_info(struct automata_list C_a) {
  struct automata_list *cap = &C_a;
  automaton aut;
  Bool check_result;
  automaton_info *aip = NULL;
  while (cap != NULL) {
    aut = automaton(*cap);
    if (check_automaton_flag) {
       check_result = aut.check_automaton();
     }
    if (aip == NULL) {
      aip = new automaton_info(aut);
    }
    else {
      aip = aip->insert(new automaton_info(aut));
    }
    cap = cap->next;
  }
  this->a = aip->a;
  this->next = aip->next;
}
