#ifndef __opoly_h__
#define __opoly_h__
#include "poly.H"


const int STRICT_CONSTANT = 2;
const int NONSTRICT_CONSTANT = 1;
const int EQUALITY_CONSTANT = 0;
const int RIGHT_RAY_CONSTANT = 0;
const int POLYHEDRON_CONSTANT = -1;


typedef lib_polyhedron poly;

matrix* nplus1_matrix(const matrix &m);
matrix* nplus1_rates_matrix(const matrix &m);
matrix* nplus1_plain_matrix(const matrix &m, int eps_pos);
matrix* my_nplus1_matrix(const matrix &m);





class opoly
{
public:
	/*
	 * Data Members
	 *
	 */

	lib_polyhedron *nplus1_form;

	                // whether nplus1_form is strict or not 
	int strict;	// if not, no mention of 0 <=eps_slack<= 1 at all
	int eps_pos;    // index of col for eps_slack var, == -1 if none
	/*
	 * Constructors
	 *
	 */
		opoly () {nplus1_form = 0; strict = 0; eps_pos = -1; }
	        // don't use these constructors for rate poly's
		// since they do automatic strict/nonstrict simplification

		opoly (matrix& constraints)
			{
			    eps_pos = constraints.nbcolumns-1;
			    nplus1_form = new lib_polyhedron(*my_nplus1_matrix(constraints),true);

			    strict = constraints.is_strict();
                            simplify();
			}

	        // consume the matrix, assumed in nplus1 form already
	        opoly  (matrix& constraints, Bool flag, int eps_adj = 0)
			{
			    eps_pos = constraints.nbcolumns-2-eps_adj;
			    nplus1_form = new lib_polyhedron(constraints,flag);
			    fix_strict();
			    simplify();
			}

		opoly (poly* p, int type = 0, int eps_adj = 0) /* Consumes p */
			{
				if (type != POLYHEDRON_CONSTANT)
					cout << "Passed wrong arg to opoly constructor\n";
				nplus1_form = p;
                                eps_pos = p->dim()-2 -eps_adj;
				fix_strict();
                                simplify();
			}

		opoly (const opoly& o) {eps_pos = o.eps_pos;
					nplus1_form = o.nplus1_form->copy();
				        strict = o.strict;
                                         }

	/*
	 * Member functions
	 *
	 */

	int 	empty();

	lib_polyhedron* closure(int return_nplus1_form=false);
	int 	contains_point(vector&);
	int 	slow_contains_point(vector&); // slower method
	int 	operator<=(opoly&);

	matrix* frame() {return nplus1_form->frame();}

	opoly*  intersection(const matrix& constraints, Bool in_nplus1_form = true) ;
	// the follwing does no simplification wrt strictness, i.e. does not
	// strip info about eps_slack (use for rates)
	opoly*  intersection_no_simp(const matrix& constraints, Bool in_nplus1_form = true) ;
	void    intersection(const opoly& o) {nplus1_form->intersection(*(o.nplus1_form->constraints()));}
	opoly*  add_elt(vector *);
	opoly*  hull(matrix frame)
	                {
			  lib_polyhedron* cl = closure();
			  cl->hull(frame);
			  opoly* the_hull = new opoly(*my_nplus1_matrix(*(cl->constraints())));
			  cl->free();
			  delete cl;
			  the_hull->strict = 0;
			  return the_hull;
			}
	void    operator=(opoly& op) {nplus1_form = op.nplus1_form;
				      strict = op.strict;} // destructive!
	opoly*	copy()
			{return new opoly(nplus1_form->copy(),POLYHEDRON_CONSTANT,dim()-2-eps_pos);}
	void	var_increment(int index) {nplus1_form->var_increment(index);}
	void	var_decrement(int index) {nplus1_form->var_decrement(index);}

	opoly* complement(); // return list of polys for compl. of single poly
	matrix*  constraints() {return nplus1_form->constraints();}

	void    invalidate_frame() {nplus1_form->invalidate_frame();}
	void    invalidate_constraints() {nplus1_form->invalidate_constraints();}
	/*
	 * Free store management
	 *
	 */
	
	void free()          
	     {
				if (nplus1_form)
					nplus1_form->free();
				 nplus1_form = 0;
	     }
	~opoly()
	     {
	       if (nplus1_form)
		 cout << "opoly destructed before freed\n";
	     }


	/*
	 * Added by HWT
	 *
	 */
	opoly *existent(int vnr, int nr_extra_vars);
	opoly *existent(int nr_exist_vars, int *indices);
	opoly *CopyColumn(int i2, int i3);
	opoly *copy_extra_to_old(int nr_keep_cols, int nr_copy_cols, 
				 int *index_p);
	opoly *vaporize_variables(int nr_v);

	void ptprint();
	void print() { ptprint(); }
	void ptprint_nplus1();
	opoly *hull(opoly *op);
	opoly *add_row(matrix *mp, int rownr, Bool negated);
	void simplify(); // test whether strict, 
	                // if nonstrict eliminate any 0<=eps<=1 constraints
        int dim();      // return dimension of the (n+1)-form
	void add_eps_constraints();
        void minimize();
        void fix_strict();
	opoly* copy_for_slopes(); // plain copy, no emptiness check

};

ostream& operator<<(ostream& , opoly& );
istream& operator>>(istream& , opoly& );


// added HWT
Bool operator == (opoly& op1, opoly& op2); 
opoly *empty_polyhedron(int nrcols);
opoly *Resize_opoly(opoly *pp4, int newsize) ;


#endif
