#include "poly.H"

// #define DEBUG_AVOID

void my_new_handler(){
   printf("Out of memory, try on a bigger machine!!\n"); exit(2);
};

Bool check_empty(matrix* F)
/* F is the frame of an empty polyhedron iff it has no vertices */
{ Bool empty = true;
  int dimension = F->nbcolumns-1;
  for(int i=0; i<F->nbrows-1; i++)
   if(F->p[i][dimension]){ empty=false; break;}
  return empty;
}

lib_polyhedron::lib_polyhedron(matrix& mat, Bool constraint)
/* 
   The constructor, from either a frame (if 'constraint' is false) 
   or a system of constraints 
*/
{

  dimension= mat.nbcolumns;
  if(constraint){
      constraints_available = true;
      frame_available = false;
      C = mat;
   }
   else{
      constraints_available = false;
      frame_available = true;
      F = mat;
   }
   empty = check_empty(this->frame());
   if(empty){ 
     // added HWT, 9/95, saves memory leak for empty poly from constraints
     invalidate_constraints();

       C = matrix(3,mat.nbcolumns);
       for(int i=0;i<F.nbcolumns-1;i++)
          { C.p[0][i]=0; C.p[1][i]=0; C.p[2][i]=0;}
       C.p[0][1]=1; C.p[0][F.nbcolumns-1]=0;  
       C.p[1][1]=1; C.p[1][F.nbcolumns-1]=-1; 
       C.p[2][0]=1; C.p[2][F.nbcolumns-1]=1;  
       constraints_available = true;
   }
} /* lib_polyhedron::lib_polyhedron */

lib_polyhedron::lib_polyhedron(lib_polyhedron& p)

{
dimension = p.dimension;
C = p.C;
F = p.F;
frame_available = p.frame_available;
constraints_available = p.constraints_available;
empty = p.empty;
};

void  lib_polyhedron::free(){
  if(constraints_available) C.free();
  if(frame_available) F.free();
  constraints_available = false;
  frame_available = false;
};


void lib_polyhedron::operator=(lib_polyhedron& p){
dimension = p.dimension;
C = p.C;
F = p.F;
frame_available = p.frame_available;
constraints_available = p.constraints_available;
empty = p.empty;
};



int 
matrix_eq(matrix *C1, matrix *C2 ) {
  int i,j;
 if ((C1->nbrows != C2->nbrows) || (C1->nbcolumns != C2->nbcolumns)) {
   cout << "Different dimensions\n";
   return 0;
 }
 else {
   int nr = C1->nbrows;
   int nc = C1->nbcolumns;
   for (i=0;i<nr;i++) {
     for (j=0;j<nc;j++) {
       if (C1->p[i][j] != C2->p[i][j]) {
	 return(0);

       }
     }
   }
   return 1;
 }
}


matrix *lib_polyhedron::constraints()
/* returns the system of constraints */
{   
    if(constraints_available) return(&C);
    else if(frame_available) { /* computation of the constraint system */
       if(empty){ 

	 // added HWT. 5/5/95
	 // deleted again, now ensure constraints_available = false
	 // => C points to no valid data at all
	 // was C.free();

         C = matrix(3,F.nbcolumns);
         for(int i=0;i<F.nbcolumns-1;i++)
            { C.p[0][i]=0; C.p[1][i]=0; C.p[2][i]=0;}
         C.p[0][1]=1; C.p[0][F.nbcolumns-1]=0;  
         C.p[1][1]=1; C.p[1][F.nbcolumns-1]=-1;  
         C.p[2][0]=1; C.p[2][F.nbcolumns-1]=1;  
	 constraints_available = true;
         return(&C);
       }
       else{

	 if (overflow_avoid_mode >=2) {
	   coeff g = F.gcd_of_frame_matrix();
	   if (g != 1) {
#ifdef DEBUG_AVOID
	     cout << "Using this GCD " << g << "\n";
	     matrix usualC;
	     transform(F,((matrix*) 0),&usualC,  true, 0);
             matrix *Fcopy = F.copy();
#endif

	   // divide frame by gcd scalar
	     F.scale_frame_matrix(g,1);
#ifdef DEBUG_AVOID
	     F.print();
#endif
	     transform(F,((matrix*) 0),&C,  true, 0); 
	     // correct for scale factor of g
	     F.scale_frame_matrix(g,0);
	     C.scale_constr_matrix(g,0);

#ifdef DEBUG_AVOID
	     cout << "fr\n";
	     F.print();
	     cout << "cons\n";
	     C.print();
             if (!matrix_eq(&F,Fcopy)) {
	       cout << "MAtrix F is altered\n";
	     }
	     Fcopy->free();
	     if (!matrix_eq(&C,&usualC)) {
	       cout << "matrices differ\n";
	       C.print();
	       cout << "the usual one is\n";
	       usualC.print();
	     }
	     usualC.free();
	     
#endif
	   }
	   else {
	     transform(F,((matrix*) 0),&C,  true, 0);
	   }
	 }
	 else {
	   // don't do anything fancy to avoid overflow
	   transform(F,((matrix*) 0),&C,  true, 0); 
	 }
         constraints_available = true;
         return(&C);
       }
     }
    else { 
       printf("Error: undefined polyhedron\n"); 
       {int *ip = NULL; *ip = 3 ; }
       exit(2);
     }
}; /* lib_polyhedron::constraints */
  
matrix *lib_polyhedron::frame()
/* returns the frame */
{
    /*
    if(empty){
       printf("Error: try to get the frame of an empty polyhedron\n");
    };
    */
    if(frame_available) return(&F);
    else if(constraints_available){ /* computation of the frame */
       transform(C, ((matrix*) 0),&F, false, 0); 
       frame_available = true;
// added HWT 9/95
       empty = check_empty(&F);
// cout << "Returning this frame";
// F.print();

       return(&F);
    }
    else{ 
       printf("Error: undefined polyhedron\n");
  {int *ip = NULL; *ip = 3 ; }
       exit(2);
    }
}; /* lib_polyhedron::frame */

lib_polyhedron *empty_lib_polyhedron(int nbcolumns)
/* return the empty polyhedron */
// fixed dangling i after first for loop, HWT 3/6/96
{ 
  int i;
  lib_polyhedron* p;
  matrix C = matrix(3,nbcolumns);
  for(i=0;i<nbcolumns-1;i++)
            { C.p[0][i]=0; C.p[1][i]=0; C.p[2][i]=0;}
  C.p[0][1]=1; C.p[0][nbcolumns-1]=0;  
  C.p[1][1]=1; C.p[1][nbcolumns-1]=-1;  
  C.p[2][0]=1; C.p[2][nbcolumns-1]=1;  
  p = new lib_polyhedron(C, true);
  p->dimension = nbcolumns;
  p->empty = true;
  p->F = matrix(1,nbcolumns);
  // HWT fixed decl of i.  3/6/96
  for(i=1;i<nbcolumns-1;i++) p->F.p[0][i]=0;
  // HWT fixed second assignment. 3/6/96.  was p[0][i]
  p->F.p[0][0]=1; p->F.p[0][nbcolumns-1]=1;
  // added HWT 9/95
  p->frame_available = true;

  return p;
};

void lib_polyhedron::print() {
   if (empty) printf("False\n"); 
   else (this->constraints())->print();
}; /* lib_polyhedron::print */

Bool operator<= (lib_polyhedron& p1, lib_polyhedron& p2)
/* is p1 included in p2? */
{
   matrix *C, *F;
   int nbcols, i, j, k;
   coeff prod;

   if(p1.empty) return(true);
   else if(p2.empty) return(false);

   /* Does the frame of p1 satisfy the constraints of p2? */
   C = p2.constraints();
   F = p1.frame();

   if(C->nbcolumns != F->nbcolumns){
       printf("Error: inconsistent polyhedron comparison\n"); exit(2);
   };
   nbcols = C->nbcolumns;
   for (i=0; i<C->nbrows-1; i++) /* iterate on constraints */
     for (j=0; j<F->nbrows-1; j++) /* iterate on elements */
     { /* compute the product */
       prod = 0;
       for (k=1; k<nbcols; k++) prod+= C->p[i][k]*F->p[j][k];
       /* does the element satisfy the constraint */
       if(C->p[i][0]){    /* i.e. an inequality */
          if(F->p[j][0])  /* i.e. a vertex or a ray */
          { if(prod<0)  { 
	    // cout << "i " << i << ", j " << j << ", prod " << prod << "\n" << flush;


                         return(false);}}
          else            /* a line */
          {  if(prod!=0) {
	    // cout << "i " << i << ", j " << j << ", prod " << prod << "\n" << flush; 
	    return(false);}}
       }
       else  /* an equality */
         if(prod != 0) return(false);
      };
      return(true);
}; /*  operator <= */

Bool operator== (lib_polyhedron& p1, lib_polyhedron& p2)
/* is p1 equal to p2 ? */
{
   return( (p1<=p2) && (p2 <= p1));
}; /* operator == */

lib_polyhedron *lib_polyhedron::var_increment (int var_index)
/* increment (on both representations) the variable of given index */
{
   int i;
 
   if(empty) return(this);

   /* Constraints: replace B by B - A[var_index] */
   if(constraints_available)
      for(i=0; i<C.nbrows; i++) 
            C.p[i][C.nbcolumns-1] -= C.p[i][var_index];

   /* Frame: increment V[var_index] by k, for each vertex V (=kW) */
   if(frame_available)
      for(i=0;i<F.nbrows-1;i++) 
          if(F.p[i][F.nbcolumns-1]) 
                  F.p[i][var_index]+= F.p[i][F.nbcolumns-1];
  return this;
}; /* lib_polyhedron::var_increment */

lib_polyhedron *lib_polyhedron::var_decrement (int var_index)
/* decrement (on both representations) the variable of given index */
{
   int i;
 
   if(empty) return(this);

   /* Constraints: replace B by B + A[var_index] */
   if(constraints_available)
      for(i=0; i<C.nbrows; i++) 
            C.p[i][C.nbcolumns-1] += C.p[i][var_index];

   /* Frame: decrement V[var_index] by k, for each vertex V (=kW) */
   if(frame_available)
      for(i=0;i<F.nbrows-1;i++) 
          if(F.p[i][F.nbcolumns-1]) 
                  F.p[i][var_index]-= F.p[i][F.nbcolumns-1];
  return this;
}; /* lib_polyhedron::var_decrement */

lib_polyhedron *lib_polyhedron::var_reset(int var_index)
/* reset to zero, on the frame,  the variable of given index */
{
  // HWT 5/9/96.  matrix *F hides lib_polyhedron's F
  // for minimal change, rename to FF, alternatively, could remove altogether
  // call frame() and reference F directly
  matrix *FF;
  int i;

  if(empty) return(this);

  FF = this->frame();
  /* reset the 'var_index' component of each generator */
  for(i=0; i<FF->nbrows-1; i++) FF->p[i][var_index]=0;

  /* compute the constraints and minimize */
  // changed HWT 9/95 to conditionally free C 
  if (constraints_available) {
    constraints_available = false;
    C.free();
  }
  (void)this->constraints();
  return this;
}; /* lib_polyhedron::var_reset */

lib_polyhedron *lib_polyhedron::var_assign(int var_index, vector *exp)
/* assign to the variable of given index, the linear expression `(ax+b)/a0' represented
   by the given vector (a b)
*/
{
  // HWT 5/96, rename F to FF
  matrix *FF;
  coeff prod;

  if(empty) return(this);
  if(exp->size != dimension){
      printf("Error: inconsistent assignment\n"); exit(2);
  }

  FF = this->frame();
/***
printf(" Assignment of var %2d with linear expression:\n", var_index);
exp->print();
printf("in the polyhedron:\n"); this->print();
***/

/* for each vertex v, replace the component of index `var_index' by `av+b' */
/* for each ray (or line) r, replace the component of index `var_index' by `ar' */
  for(int i=0;i<FF->nbrows-1;i++){
    
    prod = 0;
    for(int j=1; j<exp->size ; j++){
       prod = prod + FF->p[i][j] * exp->p[j];
       if(j !=var_index) FF->p[i][j] *= exp->p[0];
    }

    FF->p[i][var_index] = prod;
  }

  /* compute the constraints and minimize */
  // changed HWT 9/95 to conditionally free C 
  if (constraints_available) {
    constraints_available = false;
    C.free();
  }
  (void)this->constraints();
  return this;
}; /* lib_polyhedron::var_assign */

// destroys original polyhedron
lib_polyhedron *lib_polyhedron::add_elt(vector *elt)
/* Add a generating element (vertex, ray or line) */
{
/***
printf("Adjunction of the element\n"); elt->print();
printf("to the polyhedron\n"); this->print();
***/

  if(empty) return(this);
  if(elt->size != dimension){
      printf("Error: inconsistent element adjunction\n"); exit(2);
  }
  (void)this->frame();
  matrix FF(F.nbrows+1, F.nbcolumns); 
  for(int j=0; j<elt->size;j++) {FF.p[0][j]=elt->p[j];}
  // fixed i,j declarations. HWT. 3/6/96
  for(int i=0; i<F.nbrows; i++){
    for(int j=0; j<F.nbcolumns;j++) {FF.p[i+1][j] = F.p[i][j];}}
  /* added by HWT, freeing the old matrix before assigning */
  F.free();
  F = FF;
  // changed HWT 9/95 to conditionally free C 
  if (constraints_available) {
    constraints_available = false;
    C.free();
  }
  (void)this->constraints();
  return this;
}; /* lib_polyhedron::add_elt */


lib_polyhedron *lib_polyhedron::intersection(matrix c)  
/* intersects with new constraints */
{ matrix new_C;
  int nbrows, i,j;

  if(c.nbcolumns != dimension){
// force a crash
cerr << "error: inconsistent intersection\n" << flush;
cerr << "c.nbcolumns = " << c.nbcolumns << "\n" ;
cerr << "dimension = " << dimension << "\n";
cerr << flush;

    int *intp = NULL;
    *intp = 3;

      printf("Error: inconsistent intersection\n"); exit(2);
  };
  if(empty) return this;

/***
  printf("Intersection\n");
  this->print(); c.print();
***/

  
  nbrows = this->constraints()->nbrows-1;
  new_C = matrix(C.nbrows+c.nbrows-1,dimension);
  for(i=0;i<C.nbrows-1;i++) 
      for(j=0; j<dimension;j++)new_C.p[i][j] = C.p[i][j];
  for(i=0;i<c.nbrows-1;i++) 
      for(j=0; j<dimension;j++)new_C.p[nbrows+i][j] = c.p[i][j];
  for(j=0; j<dimension;j++) new_C.p[nbrows+i][j] = C.p[nbrows][j];
  C.free(); C = new_C; 
  // added HWT 9/95, conditionally free F
  if (frame_available) {
    frame_available=false;
    F.free();
  }

/***
  printf("Constraints\n"); C.print();
***/

  
  this->minimize();

/***
  printf("Frame\n"); F.print();
***/

  empty = check_empty(this->frame());

  if(empty){ 
         C.free();
         C = matrix(3,F.nbcolumns);
         for(i=0;i<F.nbcolumns-1;i++)
            { C.p[0][i]=0; C.p[1][i]=0; C.p[2][i]=0;}
         C.p[0][1]=1; C.p[0][F.nbcolumns-1]=0;  
         C.p[1][1]=1; C.p[1][F.nbcolumns-1]=-1;  
         C.p[2][0]=1; C.p[2][F.nbcolumns-1]=1;  
  }
/***
   printf("Intersection result\n"); this->print();
***/
  return this;
}; /* lib_polyhedron::intersection */

lib_polyhedron *lib_polyhedron::hull(matrix f)
/* convex union with other frame*/
{ matrix new_F;
  int nbrows, i,j;




  if(empty) return(new lib_polyhedron(f, false));
  if(f.nbcolumns != dimension){
      printf("Error: inconsistent hull\n"); exit(2);
  };

/***
  printf("Hull\n");
  printf("Frame of P1\n"); (this->frame())->print();
  printf("Frame of P2\n"); f.print();
***/

  
  nbrows = this->frame()->nbrows-1;
  new_F = matrix(F.nbrows+f.nbrows-1,dimension);
  for(i=0;i<F.nbrows-1;i++) 
      for(j=0; j<dimension;j++) new_F.p[i][j] = F.p[i][j];
  for(i=0;i<f.nbrows-1;i++) 
      for(j=0; j<dimension;j++) new_F.p[nbrows+i][j] = f.p[i][j];
   for(j=0; j<dimension;j++) new_F.p[nbrows+i][j] = F.p[nbrows][j];
  F.free(); F = new_F; 
  // added HWT 9/95, conditionally delete C, as well as setting c_av to false
  if (constraints_available) {
    constraints_available=false;
    C.free();
  }

/***
  printf("Frame of the hull\n"); F.print();
***/

  
  

  this->minimize();

/***
  printf("Result of the hull\n"); this->print();
***/


  return this;
}; /* lib_polyhedron::hull */

lib_polyhedron *lib_polyhedron::copy()
/* returns a copy of 'this', without sharing */
{ lib_polyhedron *p;
  matrix m;
  if(empty) return empty_lib_polyhedron(dimension);
  if(constraints_available){
     m = *(C.copy()); p = new lib_polyhedron(m,true);            
     return p;
  };
  if(frame_available){
     m = *(F.copy()); p = new lib_polyhedron(m,false);
     return p;
  };
  printf("Error: try to copy an undefined polyhedron\n"); exit(2);
  
}; /* lib_polyhedron::copy */


lib_polyhedron  *lib_polyhedron::minimize()
/* minimization of both representations */
{
  if(empty) return this;
  if(frame_available){
    if (constraints_available) {
      constraints_available = false;
      C.free();
    }
    (void)this->constraints();
    frame_available = false;
    F.free();
    (void)this->frame();
  }
  else if(constraints_available){
    if (frame_available) {
      frame_available = false;
      F.free();
    }
    (void)this->frame();
    constraints_available = false; 
    C.free();
    (void)this->constraints();
  }
  else{
       printf("Error: try to minimize an undefined polyhedron\n"); exit(2);
  }
  return this;
};

/* Utilitaries for widenning */

matrix *saturation(matrix *m1, matrix *m2)
/* computes the saturation matrix of m2 (frame) with respect to m1 (constraints) */
{
 int i, i1, j, k,  dimension;
 coeff P;
 Bool null;

 if(m1->nbcolumns == m2->nbcolumns) dimension= m1->nbcolumns;
 else { printf("Error: illegal matrix product\n"); exit(2);}

 matrix* prod = new matrix(m1->nbrows-1, m2->nbrows-1);
 /* computes the saturation, without considering null constraints */
 i1=0;
 for(i=0;i<m1->nbrows-1;i++){
    null = true; 
    for(j=1;j<dimension-1;j++) if(m1->p[i][j]) null =false;
    if(!null){ 
       for(k=0;k<m2->nbrows-1;k++){
          P=0;
          for(j=1;j<dimension;j++) P += m1->p[i][j] * m2->p[k][j];
          prod->p[i1][k] = (P?1:0);
       };
       i1++;
    }
 }
 prod->nbrows = i1;
 return prod;
}; /* saturation */

vector *saturation(coeff *v, matrix *m)
/* computes the saturation vector of m2 (frame) with respect to the constraint v */
{
 int j, k, dimension;
 coeff P;

 dimension= m->nbcolumns;
 vector* prod= new vector(m->nbrows-1);
 for(k=0;k<m->nbrows-1;k++){
    P=0;
    for(j=1;j<dimension;j++) P += v[j] * m->p[k][j];
    prod->p[k] = (P?1:0);
  }
 return prod;
}; /* saturation */

Bool equal(vector *v, coeff *w)
/* comparison of a vector and an array */
{
 for(int i=0;i<v->size;i++)
     if(v->p[i] != w[i]) return false; 
 return true;
}; /* equal */


Bool strictly_greater(coeff *v1, coeff *v2, int size)
/* comparison of two arrays */
{
 for(int i=0;i<size;i++) if(v1[i]>v2[i]) return true; 
 return  false;
}; /* strictly_greater */

Bool belongs(vector *v, matrix *m)
/* is v a row of m (m sorted)? */
{
 int l, r, i;
 l=0; r= m->nbrows-1;
/***
  printf("Belongs?");
  for(i=0;i<v->size;i++) printf("%2d",v->p[i]);
*/
 if(equal(v,m->p[l])) return true;
 if(equal(v,m->p[r])) return true;
 while(l<r){
    i = (l+r)/2;
    if(equal(v,m->p[i])) return true;
    if(strictly_greater(v->p,m->p[i], v->size)){
       if(l==i) break; else l=i;
    }
    else r = i;
 };
 return false;
}; /* belongs */

Bool satisfies(matrix *F, matrix *C, int i)
/* does the frame F satisfy the i-th constraint of C */
{
   int j, k;
   coeff prod;
   for (j=0; j<F->nbrows-1; j++) /* iterate on elements */
   { /* compute the product */
     prod = 0;
     for (k=1; k<F->nbcolumns; k++) prod+= C->p[i][k]*F->p[j][k];
     /* does the element satisfy the constraint */
     if(C->p[i][0]){ 
        if(F->p[j][0]) 
        { if(prod<0)  return(false);}
        else 
        {  if(prod!=0)  return(false);}
     }
     else  
       if(prod != 0) return(false);
   };
   return(true);
}; /* satisfies */

lib_polyhedron *widenning (lib_polyhedron p1, lib_polyhedron p2)
/* The standard widenning operator */
{
  lib_polyhedron *p;
  matrix mat, *F1, *F2, *C1, *C2, *sat1;
  vector *sat2; 
  coeff *v;
  int i, j, k;

  if(p1.empty) return new lib_polyhedron(p2);
  
  if(!(p1<=p2)){
     printf("Warning: the first operand of the widenning is not included in the second one\n");
     p2.hull(*(p1.frame()));
  };


  F1 = p1.frame(); C1 = p1.constraints(); 
  F2 = p2.frame(); C2 = p2.constraints();
  mat = matrix( C2->nbrows + C1->nbrows, C2->nbcolumns); k = 0;

  /* First keep the constraints of p1 satisfied by p2 */
  for(i=0;i<C1->nbrows-1; i++) if(satisfies(F2,C1,i)){
    for(j=0;j<C1->nbcolumns;j++)mat.p[k][j] = C1->p[i][j];
    k++;
  };

/***
   printf("Kept from p1\n");
   for(i=0;i<k;i++){
     for(j=0;j<C1->nbcolumns;j++) printf("%4d",mat.p[i][j]);
           printf("\n");
   };
*/
  
  sat1 = saturation(p1.constraints(), F1);
  /* sort the saturation matrix */
  for(i=1;i<sat1->nbrows;i++){
     v = sat1->p[i]; 
     j=i-1; 
     while(j>=0 && strictly_greater(sat1->p[j],v, sat1->nbcolumns)) 
                {sat1->p[j+1] = sat1->p[j]; j--;}
     sat1->p[j+1] = v;
  };

/***
   printf("Sorted saturation\n");
   for(i=0;i<sat1->nbrows;i++){
      for(j=0;j<sat1->nbcolumns;j++) printf("%2d",sat1->p[i][j]);
      printf("\n");
   };
*/
   
  for(i=0;i<C2->nbrows-1;i++){
     sat2 = saturation(C2->p[i],F1);
     if(belongs(sat2,sat1)){
/***
         printf("Yes\n");
*/
        for(j=0;j<C2->nbcolumns;j++)mat.p[k][j] = C2->p[i][j];
        k++;
     }
  }
  /* Last row */
  for(j=1;j<C2->nbcolumns-1;j++)mat.p[k][j] =0;
  mat.p[k][0]=1; mat.p[k][j] = 1;

  mat.nbrows = k+1;
  p = new lib_polyhedron(mat, true);

  (void)p->frame();
  
  return p;
}; /* widenning */

lib_polyhedron *limited_widenning (lib_polyhedron p1, lib_polyhedron p2, matrix cont)
/* widenning upto given constraints : the standard widenning is
   intersected with the given constraints satisfied by the widenner
*/
{ 
  lib_polyhedron *p;
  matrix sat_cont, *F;
  int i,j,k,l;
  coeff prod;
  Bool satisfied;

  
  

  if(p1.empty) return new lib_polyhedron(p2);

  if(!(p1<=p2)){
     p2.hull(*(p1.frame()));
  };

  p = widenning(p1,p2);

  
  

  /* select the constraints satisfied by p2 */
  sat_cont = matrix(cont.nbrows+1, cont.nbcolumns);
  F = p2.frame(); l=0;
  
  for(i=0;i<cont.nbrows;i++){
     
     satisfied = 1;
     for(k=0;k<F->nbrows-1;k++){
         prod=0;
         for(j=1;j<F->nbcolumns;j++) prod+= cont.p[i][j]*F->p[k][j];
         /* does the element satisfy the constraint */
         if(cont.p[i][0]){ 
             if(F->p[k][0]){ 
                   if(prod<0) {satisfied = false; break;}
             }
             else {
                  if(prod!=0)  {satisfied = false; break;}
             }
         }
         else { 
             if(prod!=0) {satisfied = false; break;}
         }
     };
     if(satisfied) 
         sat_cont.p[l++]=cont.p[i];
  };
  /* Last row */
  for(j=0;j<sat_cont.nbcolumns-1;j++) sat_cont.p[l][j] = 0;
  sat_cont.p[l][0]=1; sat_cont.p[l][j]=1;
  sat_cont.nbrows=l+1;

  
  

  /* intersect */
  p = p->intersection(sat_cont); 

  return p;
};

lib_polyhedron *bounded_widenning (lib_polyhedron p1, lib_polyhedron p2, matrix cont)
/* similar to limited_widenning, but preserves also intervals:
   any constraint k<x<k' satisfied by both arguments will be preserved
*/
{ int i,j,k;
  if(p1.empty) return new lib_polyhedron(p2);

  matrix cont1 = matrix(cont.nbrows+ 2*(cont.nbcolumns-2),  cont.nbcolumns);
  for(i=0;i<cont.nbrows-1;i++) cont1.p[i]=cont.p[i];

  
  
  matrix F = *(p1.frame());
  int lastcol = F.nbcolumns-1;
  
  int fv;
  for(fv=0; fv < F.nbrows-1; fv++){if(F.p[fv][0] && F.p[fv][lastcol]) break;}
  if(fv == F.nbrows-1){printf("Error: widenned has no vertex\n"); exit(2);}

  coeff inf, infcoef, sup, supcoef;
  Bool infbounded, supbounded;
  for(k=1;k<F.nbcolumns-1;k++){ 
    inf = F.p[fv][k]; sup = F.p[fv][k]; 
    infcoef = F.p[fv][lastcol]; supcoef = F.p[fv][lastcol];
    for(j=fv+1; j<F.nbrows-1;j++){ 
      if(F.p[j][0] && F.p[j][lastcol]){
        if(F.p[j][k]*infcoef < inf*F.p[j][lastcol]){inf=F.p[j][k];infcoef=F.p[j][lastcol];}
        else 
        if(F.p[j][k]*supcoef > sup*F.p[j][lastcol]){sup=F.p[j][k];supcoef=F.p[j][lastcol];}
      }
    }
    
    infbounded = true; supbounded = true;
    for(j=0;j<F.nbrows-1;j++){ 
       if(F.p[j][lastcol]==0){
         if(F.p[j][k]>0) supbounded=false;
         else if (F.p[j][k]<0) infbounded=false;
       }
    }
    
    if(infbounded){
      cont1.p[i][0]=1; for(j=1;j<lastcol;j++)  cont1.p[i][j]=0;
      cont1.p[i][k]=infcoef; cont1.p[i][lastcol]=-inf; i++;
    }
    if(supbounded){
      cont1.p[i][0]=1; for(j=1;j<lastcol;j++)  cont1.p[i][j]=0;
      cont1.p[i][k]=-supcoef; cont1.p[i][lastcol]=sup; i++;
    }
  }

  
  cont1.p[i]=cont.p[cont.nbrows-1];
  cont1.nbrows=i+1;

  

  return limited_widenning(p1,p2,cont1);

}
 

lib_polyhedron *lib_polyhedron::add_a_null_variable(){

   int i, j, nbrows, nbcolumns;

   if(empty) return empty_lib_polyhedron(dimension+1);
   this->constraints();
   nbrows = C.nbrows;
   nbcolumns = C.nbcolumns;
   matrix m= matrix(nbrows+1,nbcolumns+1);
   for(i=0;i<nbrows-1;i++){
     for(j=0;j<nbcolumns-1;j++) m.p[i][j]=C.p[i][j];
     m.p[i][j]=0; m.p[i][j+1]=C.p[i][j];
   };
   /* one more equation */
   for(j=0;j<nbcolumns+1;j++) m.p[i][j]=0;
   m.p[i][nbcolumns-1]=1;
   /* the last row */
   m.p[++i][0]=1; m.p[i][nbcolumns]=1;
    for(j=1;j<nbcolumns;j++) m.p[i][j]=0;
   
   return new lib_polyhedron(m,true);
       
};


lib_polyhedron *lib_polyhedron::delete_variable(int d){
/* note, it returns a value, but does not itself alter the object! */
/* also, it suppresses the variable, it does existential quant on var */
/* via this->frame() before removing it */
   int i, j, nbrows, nbcolumns;

  this->frame();
  nbrows = F.nbrows;
  nbcolumns = F.nbcolumns;
  if(d<1 || d>=nbcolumns){
     printf("Error: try to delete a variable out of the domain\n");
     exit(2);
  };
  matrix m = matrix(nbrows,nbcolumns-1); 
  for(i=0;i<nbrows;i++){
    for(j=0;j<d;j++) m.p[i][j]=F.p[i][j];
    for(j=d+1;j<nbcolumns;j++) m.p[i][j-1]=F.p[i][j];
  };
   return new lib_polyhedron(m,false);
};

/*
------------------------- original version has a bug in it --------------
lib_polyhedron *lib_polyhedron::delete_variable(int d){

   int i, j, nr_rows, nbcolumns;

  this->F = *(this->frame());
  nr_rows = F.nbrows;
  nbcolumns = F.nbcolumns;
  if(d<1 || d>=nbcolumns){
     printf("Error: try to delete a variable out of the domain\n");
     exit(2);
  };
  matrix m = matrix(nr_rows,nbcolumns-1); 
  for(i=0;i<nr_rows;i++){
    for(j=0;j<d;j++) m.p[i][j]=F.p[i][j];
    for(j=d+1;j<nbcolumns;j++) m.p[i][j-1]=F.p[i][j];
  };
printf("New number columns should be %d\n", nbcolumns-1);
printf("m has %d columns\n", m.nbcolumns);
   return new lib_polyhedron(m,false);
};
// ------------------------------------------------------------
*/ 
