/* header file for parser in predicates (in C) */

#ifndef __PREDICATE_H
#define __PREDICATE_H

/* use #define __DEBUG_PRED to get output on predicate parsing */

#include "include.h"
#include "output.h"
#include "basic.h"

#ifdef __cplusplus
extern "C" {
#endif

/* a state predicate */
/* i.e. a node for a conjunctive list of requirements */
struct state_predicate {
  char *aut_str;   /* identifies the name of the automaton */
  char *locn_str;  /* identifies locn within that automaton */
  struct state_predicate *next;
};

/* a convex predicate */
/* i.e. a node containing a conjunctive list of state_predicate requirements */
/* and a ptr to a _p_matrix for the constraints */
/* N.B. a list of these convex pred's should be interpreted as disjunctive */
struct convex_predicate {
  struct state_predicate *state_pred; /* ptr to conjunctive list */
  struct _p_matrix       *var_pred;

  struct convex_predicate *next;  /* ptr to next on disjunctive list */
};


/* ============= routines for parsing ============ */

/* operate on the following globals: 
 
final_pred: the resulting predicate (non-convex) for the entire predicate
convex_pred:       the current convex predicate
p_state_pred:      the currently parsed state predicate
vect_head: try to follow parser code for these variables use as
                       it appears elsewhere in the code
*/

extern struct convex_predicate *final_pred;
extern struct convex_predicate *convex_pred;
extern struct state_predicate *p_state_pred;

extern char *state_aut_name;

void
print_convex_predicate(struct convex_predicate *cp);

void init_pred_spec();
#ifdef __cplusplus
}
#endif


#endif
