#include "poly.H"
#include "extrapoly.H"
#include "automaton.H"
#include "region.H"
#include "include.h"
#include "convert.H"
#include "predicate.h"

// #define DEBUG_VERIF

// consider adding points in traces, instead of outputting sets.
//
// assume from forward reachability
// call set_pre()
// choose final point in reached & final, p_n say, in R_n
// find pre_sigma(pre_tau(p_n)), call this region R_{n-1}
// find point p_{n-1} in in R_{n-1}
// find intermediate point p'_n in post_sigma(p_{n-1}) ^ pre_tau(p_n)

// use the eps_slack variable as  a clock timer for the time elapsed.



// insert regp, the region pted to (a single region, not a list), into the list
// pted to by reglistp
// no need to make a copy, can use this actual region
// works fine if *reglistp == NULL, but not is regp is NULL;
void
insertreg(region *regp, region **reglistp) {
  region * temp = *reglistp;
  // add to head of existing list.
  *reglistp = regp;
  regp->next = temp;
}




// compute OR of the list of regions.
// add regions in regp2 to those in *reglist
// the original *reglist may be lost in computing the new list, but keep 
//   all data from regp2 intact
// HWT: resulting list not previously simplified correctly. fixed 3/28.
//      prev. regp started at *prevp instead of *reglistaddr
void
region_or(region **reglistaddr, region *regtoadd) {
  region *regp2 = regtoadd;
    // for each region in list to add ...
    while (regp2 != NULL) {
      region **prevp = reglistaddr;
      region *regp = *prevp;
      Bool contained_in = false;
      // check wrt current added list.
      while ((regp != NULL) && !contained_in) {
	contained_in = ((*regp2) <= (*regp));
	if (!contained_in) {
	  // regp2 must still be considered as being new (so far)
	  // but test if regp should be deleted
	  if ( (*regp) <= (*regp2)) {
	    // avoid memory leak
	    // skip over (and delete) this element
	    region *tempr = regp;
	    *prevp = regp->next;
	    tempr->free();
	    regp = *prevp;
	  }
	  else {
	    // need to keep the (*regp) link in original list
	    // merely update ptrs
	    prevp = &(regp->next);
	  regp = regp->next;
	  }
	}
      }
      if (!contained_in && !regp2->is_empty()) {
	insertreg(regp2->fresh_copy(), reglistaddr);
      }
      regp2 = regp2->next;
    }
}


//  --------------------------------------------------


void
region::print_0(Bool print_name, Bool print_list) {
  single_print_0(print_name);
  if (print_list && (next != NULL)) {
    // don't print the name anymore, do continue printing the list
    next->print_0(false,print_list);
  }
}

void
region::single_full_print() {
  cout << "Reg ID: " << reg_id << "\n";
  if (trans_in == NULL) {
    cout << "Trans in: ---\n";
  }
  else {
    trans_in->brief_print();
  }
  cout << "Pred. reg id: " << pred_reg_id << "\n";

  locnp->name_print();
/*   polyp->print(); */
  polyp->ptprint();
  

/* (This doesn't check for emptiness)  polyp->constraints()->ptprint();  */

}


void
region::single_print_0( Bool print_name_flag) {


  if (print_name_flag) {
    locnp->name_print();
  }
  else {
    cout << "| \n";
  }
//  polyp->print();
  polyp->ptprint();

/* (This doesn't check for emptiness)  polyp->constraints()->ptprint();  */

}


void
region::single_print() {
  single_print_0(true);
}

void
region::print() {

  single_print();
  if (next != NULL) {
    next->print();
  }
}


void
region::minimize() {
  if (polyp != NULL) {
    polyp->minimize();
  }
  if (next != NULL) {
    next->minimize();
  }
}

void
region::min_print() {
  polyp->minimize();
  single_print();
  if (next != NULL) {
    cout << "  |\n";
    next->min_print();
  }
}

// print as above, but without the repeated locn names for the list
void
region::list_print() {
  single_print();
  if (next != NULL) {
    // print without the name, but the entire list
    next->print_0(false,true);
  }
}

int
region::list_size() {
  region * regp = this;
  int listsize = 0;
  while (regp != NULL) {
    listsize++;
    regp = regp->next;
  }
  return(listsize);
}


region::region() {
  locnp = NULL;
  polyp = NULL;
  reg_id = -1;
  pred_reg_id = -1;
  trans_in = NULL;
  next = NULL;
}

region::region(region& r) {
  locnp = r.locnp;
  polyp = r.polyp;
  reg_id = r.reg_id;
  pred_reg_id = r.pred_reg_id;
  trans_in = r.trans_in;
  next = NULL;
}


region::region(location *locp, polyhedron *pp) {
  locnp = locp;
  polyp = pp;
  reg_id = -1;
  pred_reg_id = -1;
  trans_in = NULL;
  next = NULL;
}

region *
region::copy() {
// has pointer to same locations and poly, not pointer to copies of locn/poly.
//
  region *temp = new region;
  temp->locnp = locnp;
  temp->polyp = polyp;
  // should this be NULL instead?
  temp->reg_id = reg_id;
  temp->pred_reg_id = pred_reg_id;
  temp->trans_in = trans_in;
  temp->next = next;
  return temp;
}

region *
region::fresh_copy() {
// has pointer to same location and COPY of poly, not pointer to same poly
//
  region *temp = new region;
  temp->locnp = locnp;
  temp->polyp = polyp->copy();
  temp->next = NULL;
  /* just copy the trans ptr, not the data */
  temp->reg_id = reg_id;
  temp->pred_reg_id = pred_reg_id;
  temp->trans_in = trans_in;
  return temp;
}


region *
region::fresh_copy(int new_reg_id) {
// has pointer to same location and COPY of poly, not pointer to same poly
//
  region *temp = fresh_copy();
  temp->reg_id = new_reg_id;
  return temp;
}


region *
region::fresh_copy_list() {
  region *listp = NULL;  // the copied list
  region * regp = this;  // for iterating along list
  while (regp != NULL) {
    insertreg( regp->fresh_copy(),&listp);
    regp = regp->next;
  }
  return(listp);
}

region *
region::insert(region * r) {
//
// insert r at head of list
// return new list head
// assumes only have single item to insert!
  if (r != NULL) {
    r->next = this;
    return(r);
  }
  else {
    return( this);
  }
}

Bool
region::is_contained_in(region& r) {
//
// test to see if region is contained in r
//
  return( (locnp == r.locnp) && (*polyp <= *(r.polyp)));
}

Bool
region::is_equal(region & r) {
//
  return( (locnp == r.locnp) && (*polyp == *(r.polyp)));
}

Bool
region::is_empty() {
//
  if (polyp == NULL) {
    return(true);
  } 
  else {
    return( polyp->empty() );
  }
}

int
list_length(region *rp) {
  int leng = 0;
  while (rp) {
    cout << "." << flush;
    if (rp->polyp)
      rp = rp->next;
    else {
      rp = NULL;
    }
    cout << "+" << flush;
    leng++;
  }
  return leng;
}


// is the list pted to all empty poly's?
// i.e. return 1 if there exists a non-empty poly on list pted to
Bool
region::is_empty_list() {
  if ((polyp == NULL) || (polyp->empty())) {
    if (next != NULL) {
      return (next->is_empty_list());
    }
    else {
      return(true);
    }
  }
  else {
    return(false);
  }
}

void
region::free() {
  if (polyp) {
    polyp->free();
  }
  delete this;
}

// free the entire list pted to by this region
void
region::free_list() {
  if (next != NULL) {
    next->free_list();
  }
  free();
}


// both these routines return a fresh copy of data
// copy first, since the poly routine destroys the data.
// potential memory leak: if ptr not kept before intersection is called
region *
region::intersect(polyhedron *pp) {
  region *newreg = fresh_copy();

  newreg->polyp->intersection(*(pp->constraints()));
  return(newreg);
}


// return fresh region for the intersection
// 9/19/95: N.B. must leave original region, and rp unchanged.
region *
region::intersect(region *rp) {
  if (rp == NULL) {
    return (fresh_copy());
  }
  if (strcmp(locnp->name,rp->locnp->name) == 0) {
    region *result = intersect(rp->polyp);
    return(result);
  }
  else {
    region *newreg = copy();
    newreg->polyp = empty_polyhedron(var_count+2);
    return( newreg );
  }
}


// simplify a list of regions, using region_or
// doesn't destroy the original list
region *
region::simplify() {
    region *rp = (region *) NULL;
    region_or(&rp,this);
    return(rp);
}

Bool
operator <= (region r1, region r2) {
  return( (strcmp(r1.locnp->name,r2.locnp->name) == 0) &&
	  (*(r1.polyp) <= *(r2.polyp)) );
}


// assuming non-empty region, etc. change the name so locn of 
// process proc_nr has blank name "".
// nasty since changes the actual name of the location in the automaton
void
region::elim_loc_name_for_proc( int proc_nr) {
  int i;
  char *front = locnp->name;
  char *back, *newname;
  // find ptr to char immediately after proc_nr'th "."
  for (i=0;i<proc_nr;i++) {
    front = strchr(front,'.');
    front++;
  }
  // find trailing end of this locn's name
  back = front;
  while ( (*back != '\0') && (*back != '.')) {
    back++;
  }
  newname = new char[strlen(locnp->name) - (back-front) + 1];
  // copy everything in front of 'front', i.e. up to last '.'
  strncpy(newname, locnp->name, front-locnp->name);
  // copy rest of string into end.
  char *p = newname + (front-locnp->name);
  char *q = back;
  while (*q != 0) {
    *p++ = *q++;
  }
  *p = *q;
  delete (locnp->name);
  locnp->name = newname;
}

// ======  begin closure difference routines ========================


// add constraint from row rownr of rp's
// constraint matrix to object's constraints
// assumes the constraint being added is an inequality
// assumes locn's match
// destroy original object, return ptr to result
region * 
region::add_row(region *rp, int rownr) {
  // last parameter causes row to be added without negation
  region * result = 
    new region(locnp, polyp->add_row(rp->polyp->constraints(), rownr, false));
  free();
  return(result);
}




// ======== end closure difference routines ========================


// returns exactly whether object contained in rp's list
Bool 
region::exact_is_in(region * rp) {
  Bool contained_in;
  region *diff = minus(rp);
  if (diff) {
    contained_in  = diff->is_empty_list();
    diff->free_list(); 
  }
  else {
    contained_in = true;
  }
  return (contained_in);
} 


// returns exactly whether states in object's list of regions are equivalent
// to states in rp's list of regions
Bool 
region::exact_is_equal(region *rp) {
  return( exact_is_in(rp) && rp->exact_is_in(this) );
}


//   for generating error traces


// assign new IDs to entire list
// use global max_reg_id
void 
region::assign_reg_ids() {
  max_reg_id++;
  reg_id = max_reg_id;
  if (next != NULL) {
    next->assign_reg_ids();
  }
}

Bool
region::is_non_strict() {
  return( ! (polyp->strict) );
}

//  **********************************************************
// for region arrays




region_array::region_array() {
    size = 0;
    regions = (region **) NULL;
    init_regn = (region_array *) NULL;
    autp = (automaton *) NULL;

}

region_array::region_array(int sz) {
  int i;
  size = sz;
  regions = (region **) malloc (sz * sizeof(region *));
  for (i=0;i<sz;i++) {
    regions[i] = (region *) NULL;
  }
  init_regn = (region_array *) NULL;
  autp = (automaton *) NULL;
}

// added 12/96
region_array::region_array(int sz, automaton *automaton_p) {
  int i;
  size = sz;
  regions = (region **) malloc (sz * sizeof(region *));
  for (i=0;i<sz;i++) {
    regions[i] = (region *) NULL;
  }
  init_regn = (region_array *) NULL;
  autp = automaton_p;
}

// changed 7/27/95 to use region constructor for inserting new region
region_array::region_array(region_array *rega) {
  int i;
  size = rega->size;
  regions = (region **) malloc (size * sizeof(region *));
  for (i=0;i<size;i++) {
    region *rp = rega->regions[i];
    regions[i] = (region *) NULL;
    while (rp != NULL) { // FIX THIS. the NULL poly's
      if (rp->polyp != NULL) {
//	insertreg(new region(rp->locnp,rp->polyp->copy()), &(regions[i]));
	insertreg(rp->fresh_copy(), &(regions[i]));
      }
      rp = rp->next;
    }
  }
  if (rega->init_regn != NULL) {
    init_regn = new region_array(rega->init_regn);
    init_regn->autp = rega->autp;
  }
  else {
    init_regn = (region_array *) NULL;
  }
  autp = rega->autp;
}



// added 2/96
region_array::region_array(int sz, region *regp) {
  int i;
  size = sz;
  regions = (region **) malloc (size * sizeof(region *));
  for (i=0;i<size;i++) {
    regions[i] = (region *) NULL;
  }
  region *rp = regp;
  while (rp != NULL) { 
    if (rp->polyp != NULL) {
      insertreg(rp->fresh_copy(), &(regions[rp->locnp->locn_id]));
    }
    rp = rp->next;
  }
  init_regn = (region_array *) NULL;
  autp = (automaton *) NULL;
}

// yields whether every polyp points to a nonstrict region
Bool
region_array::ra_is_non_strict() {
  int i;
  for (i=0;i<size;i++) {
    region * regp = regions[i];
    while (regp != NULL) {
      if (regp->polyp != NULL) {
	if (!regp->is_non_strict()) {
	  return( false);
	}
      }
      regp = regp->next;
    }
  }
  return(true);
}

int
automaton_index(automaton *autp, struct state_predicate *sp) { 
  Bool found = false;
  int aut_pos = 0;
  while  (!found && (aut_pos <autp->nr_components)) {
    found = (strcmp(sp->aut_str,
		    autp->comp_inf_p[aut_pos].automaton_name) == 0);
    if (!found) {
      aut_pos++;
    }
  }
  if (!found) {
    cout << "Error: automaton name (" << sp->aut_str << 
      ") not found\n" << flush;
    exit(1);
  }
  else {
    return(aut_pos);
  }
}
 

// return the last 2 params: indices for every automaton name, and nr conjuncts
void
find_automata_indices(automaton *autp,struct convex_predicate *cp,
		      int **aut_pos_p, int *nr_conj_p) {
  struct state_predicate *sp = cp->state_pred;
  (*nr_conj_p) = 0;
  while (sp != NULL) {
    if (*nr_conj_p == 0) {
      *aut_pos_p = (int *) malloc (sizeof(int));
    }
    else {
      *aut_pos_p = (int *) realloc((*aut_pos_p), (*nr_conj_p+1)*sizeof(int));
    }
    (*aut_pos_p)[*nr_conj_p] = automaton_index(autp,sp);
    (*nr_conj_p)++;
    sp = sp->next;
  }
}


// check whether the locn name exists in the aut_pos'th automaton component
void
check_loc_name_exists(automaton *autp, struct state_predicate *sp,
		       int aut_pos) {
  Bool found = false;
  int i = 0;
  while (!found && (i<autp->comp_inf_p[aut_pos].nr_locations)) {
    found = (strcmp(sp->locn_str,
		    autp->comp_inf_p[aut_pos].locn_names[i])==0);
    if (!found) {
      i++;
    }
  }
  if (!found) {
    cout << "Error: location name (" << sp->locn_str << 
      ") does not exists in "
      << "automaton " << sp->aut_str << "\n" << flush;
    exit(1);
  }
}


void
check_loc_names_exist(automaton *autp, struct convex_predicate *cp,
		       int *aut_pos, int nr_conj) {
  int i;
  struct state_predicate *sp = cp->state_pred;
  for (i=0;i<nr_conj;i++) {
    check_loc_name_exists(autp,sp,aut_pos[i]);
    sp = sp->next;
  }
}



// return whether this location matches the state_predicate (1 conjunct)
// given that the locn belongs to the aut_pos'th automaton in composition
// skips past the "." used to separate locations -
Bool
matches_locn_name(struct state_predicate *sp, int aut_pos,
		  char * prod_loc_name) {
  int i;
  char * rest_of_loc_name = prod_loc_name;
  for (i=0;i<aut_pos;i++) {
    rest_of_loc_name = strstr(rest_of_loc_name,".");
    rest_of_loc_name++;
  }
  if (strstr(rest_of_loc_name,sp->locn_str) == rest_of_loc_name) {
    // matches the prefix of the string for the locn in the correct position
    // check for end of loc name, or '.' immediately following
    return( (strlen(rest_of_loc_name) == strlen(sp->locn_str))
	    || rest_of_loc_name[strlen(sp->locn_str)] == '.' );
  }
  else {
    // prefix doesn't even match
    return(false);
  }
}


// return whether this location matches the convex_predicate
// given that the locn belongs to the aut_pos'th automaton in composition
// skips past the "." used to separate locations
// there are nr_conj state_pred's in the convex predicate
Bool
matches_locn_names(struct convex_predicate *cp, int *aut_pos_p, int nr_conjs,
		  char * prod_loc_name) {
  Bool matches = true;
  struct state_predicate *sp = cp->state_pred;
  // i is the number of conjs we have matched so far
  int i=0;
  while (matches && (i<nr_conjs)) {
    matches = matches_locn_name(sp,aut_pos_p[i], prod_loc_name);
    if (matches) {
      i++;
      sp = sp->next;
    }
  }
  return(i==nr_conjs);
}


// return ptr to region array for the single predicate cp, i.e. don't
// worry about the rest of the list.
//  autp is the reference automaton. 
// do some checking at same time that state names and locn names are valid.
region_array *
ra_for_single_pred(automaton *automaton_p, struct convex_predicate *cp) {
  if (cp==NULL) {
    cout << "Error: called region_array constructor with empty predicate\n";
    cout << flush;
    exit (1);
  }
  else {
    int i;
    int *aut_pos;  // array of indices for automata indices 
    int nr_state_conj; // nr of conjuncts in state predicate
    region_array *rap = 
      new region_array(automaton_p->nr_locations,automaton_p);
    matrix m = matrix(cp->var_pred->rows, cp->var_pred->columns);
    Bool found = false;
    Cpp_matrix_nplus1(&m, cp->var_pred);
    // find automaton name
    find_automata_indices(automaton_p,cp, &aut_pos, &nr_state_conj);

    // check locn names exist
    // exits program if it does not
    check_loc_names_exist(automaton_p,cp,aut_pos,nr_state_conj);

    for (i=0;i<automaton_p->nr_locations;i++) {
      // check whether name is here
      if (matches_locn_names(cp,aut_pos,nr_state_conj,automaton_p->locations[i].name)) {
	// add copy?  of matrix m to otherwise NULL list for the locn
	rap->regions[i] = new region(&(automaton_p->locations[i]),
				 new polyhedron(*(m.copy()),true));
      }
    }
    rap->autp = automaton_p;
    return(rap);
  }
}


// construct a region array corresponding to the disjunctive predicate
// of convex parts.
// autp is the reference automaton.  it's a product automaton.
region_array *
ra_for_pred(automaton *automaton_p, struct convex_predicate *cp) {
  if (cp==NULL) {
    cout << "Error: called region_array constructor with empty predicate\n";
    cout << flush;
    exit (1);
  }
  else {
    if (cp->next == NULL) {
      return ( ra_for_single_pred(automaton_p,cp) );
    }
    else {
      // cp->next != NULL
      region_array *rap = ra_for_single_pred(automaton_p,cp);
      region_array *raprest = ra_for_pred(automaton_p,cp->next);
      raprest->autp = automaton_p;
      return( raprest->ra_or(rap) );
    }
  }

}

region_array::region_array(struct convex_predicate *cp, automaton *aut) {
  *this = *ra_for_pred(aut,cp);
}

void
region_array::check_sizes_match(region_array ra2, char *routinename) {
  if (size != ra2.size) {
    cout << "Error in call to routine " << routinename << "\n";
    cout << "Sizes don't match\n" << flush;
    exit(1);
  }
}

void
global_free_region(region ** ptr) {
  free (ptr);
}


void
region_array::free() {
  int i;
  for (i=0;i<size;i++) {
    region *regp = regions[i];
    while (regp != NULL) {
      region *nextreg = regp->next;
      regp->free();
      regp=nextreg;
    }
  }
  global_free_region(regions);
  if (init_regn != NULL) {
    init_regn->free();
  }
}

Bool
region_array::is_empty() {
  int i=0;
  Bool empty = true;
  while ((i<size) && empty) {
    // 9/27/96. was "empty = (regions[i]==NULL)"
    empty = (regions[i] == NULL || regions[i]->is_empty_list() );
    i++;
  }
 return(empty);
}




// add a list of regions to the region array
// destroy the list rp in the process
// blindly adds all elements to the list, without care for redundancies
void
region_array::add_to(region *rp) {
  region *regp = rp;
  region *nextp;
  while (regp != NULL) {
    nextp = regp->next;
    if ((regp->locnp->locn_id > size) ||
	(regp->locnp->locn_id < 0)) {
      cout << "Illegal locn id in region_array::add_to()\n" << flush;
      exit(1);
    }
    insertreg(regp,&(regions[regp->locnp->locn_id]));
    regp = nextp;
  }
}


// yields whether the object is "in" its argument
// containment defined as each region in object lies in some region in arg
Bool
region_array::ra_in(region_array *rap) {
  Bool all_in = true;
  Bool this_in; 
  int i = 0;
  region * regp, *regp2;

  check_sizes_match(rap, "ra_in");
  while (all_in && (i<size)) {
    // check each region with locn i
    regp = regions[i];
    while (all_in && (regp != NULL)) {
      if (! (regp->polyp->empty()) ) {
	// check all possible containing regions regp2
	this_in = false;
	regp2 = rap->regions[i];
	while (!this_in && (regp2 != NULL) ) {
	  this_in = (*regp <= *regp2);
	  if (!this_in) {
	    regp2 = regp2->next;
	  }
	}
	all_in = this_in;
      }
      if (all_in) {
	regp = regp->next;
      }
    }
    if (all_in) {
      i++;
    }
  }
  return(all_in);
}

// returns whether the region array is non-empty
Bool
region_array::ra_empty() {
  int i =0;
  region *regp;
  Bool empty = true;
  while ((i<size) && empty) {
    regp = regions[i];
    while (empty && (regp != NULL)) {
      empty = regp->is_empty();
      if (empty) {
	regp = regp->next;
      }
    }
    if (empty) i++;
  }
  return (empty);
}

// yields whether the object is "equivalent" to the argument
// equivalence is defined as each region in one lies in some region in other
Bool
region_array::ra_equiv(region_array *rap) {
  return( ra_in(rap) && rap->ra_in(this) );
}

// altered HWT 5/23/95 to allow full post, full pre
// flag fullsucc = 1 for full, 0 for old version, without actual region's
//      cts evolution explicitly added
region_array *
region_array::ra_succ(enum directions dirn, int fullsucc) {
  // go through whole list of locations
  // for each go through each item on list
  int i;
  region *succp;
  region *rp;
  region_array *succ_ra = new region_array(size, autp);
  for (i=0;i<size;i++) {
    rp = regions[i];
    while (rp != NULL) {
      if (dirn == PRE) {
	if (fullsucc) {
	  succp = rp->full_pre();
	}
	else {
	  succp = rp->pre();
	}
      }
      else {
	// doing post
	if (fullsucc) {
	  succp = rp->full_post();
	}
	else {
	  succp = rp->post();
	}
      }
      succ_ra->add_to(succp);
      rp = rp->next;
    }
  }
  return(succ_ra);
}


region_array *
region_array::ra_post() {
  return(ra_succ(POST,0));
}


region_array *
region_array::ra_pre() {
  return(ra_succ(PRE,0));
}


region_array *
region_array::ra_full_post() {
  return(ra_succ(POST,1));
}


region_array *
region_array::ra_full_pre() {
  return(ra_succ(PRE,1));
}



// compute the intersection of the two region arrays
// assume the locations line up by index, etc.
// does NOT destroy the old region_array (this), nor ra2.
//
region_array *
region_array::ra_and(region_array ra2) {
  // algorithm creates a temp array with all the intersected pieces
  //     by intersecting all pairs
  // then use ra_or with an empty array to get the result

  int i;
  region_array *result;

  region_array tempra = region_array(size,autp);
  check_sizes_match(ra2,"ra_and");
  for (i=0;i<size;i++) {
    region *regp = regions[i];
    while (regp != NULL) {
      if (!(regp->is_empty())) {
	region *regp2 = ra2.regions[i];
	while (regp2 != NULL) {
	  // the intersected region
	  region *intreg = regp->intersect(regp2);
	  if (!(intreg->is_empty())) {
	    // if did reach with edges, then may be about to generate a trace
	    if (generating_trace && did_reach_with_edges) {
	      // identify the 'parent' enclosing reached region
	      // for backtracking in trace generation
	      intreg->reg_id = regp->reg_id;
	    }
	    insertreg(intreg, &(tempra.regions[i]));
	  }
	  else {
	    intreg->free();
	  }

	  regp2 = regp2->next;
	}
      }
      regp = regp->next;
    }
  }
  // ra_or returns ptr to emptyra (calling object)
  if (!generating_trace) {
    region_array *emptyra = new region_array(size, autp);
    result = emptyra->ra_or(tempra);
  }
  else {
    // avoid the ra_or that joins containing regions, since it results
    // in longer paths
    result = new region_array(&tempra);
  }
  tempra.free();
  result->autp = autp;
  return(result);
}


/*
// compute OR of the list of regions.
// add regions in regp2 to those in *reglist
// the original *reglist may be lost in computing the new list, but keep 
//   all data from regp2 intact
void
region_or(region **reglistaddr, region *regp2) {
  region **prevp = reglistaddr;
    while (regp2 != NULL) {
      region *regp = *reglistaddr;
      Bool contained_in = false;
      while ((regp != NULL) && !contained_in) {
	contained_in = ((*regp2) <= (*regp));
	if (!contained_in) {
	  regp = regp->next;
	}
      }
      if (!contained_in) {
	insertreg(regp2->fresh_copy(), reglistaddr);
      }
      regp2 = regp2->next;
    }
}
*/

// compute the union of the two region arrays
// assume the locations line up by index, etc.
// can destroy the old region_array (this), but not ra2.
// [thus when calling it, if we want the old version, must save first].
//
// try adding second operator to this.
// no merging of regions, no containment detected unless entirely
// within a region, not detected if within two regions, but no single region
region_array *
region_array::ra_or(region_array ra2) {
  int i;
 
  check_sizes_match(ra2,"ra_or");
  
  for (i=0;i<size;i++) {
    region *regp2 = ra2.regions[i];
    region_or(&regions[i],regp2);
  }
  return(this);
}


// compute the union of the two region arrays
// assume the locations line up by index, etc.
// can destroy the old region_array (this), which gets returned as
// the answer, but not ra2.
// [thus when calling it, if we want the old version, must save first].
// also store in ranew the regions that are "new", i.e. not contained
// in this's states.  (For now just do set containment, not per elt
// within each region).
//
// only differs from ra_or in one main line, i.e. insert reg into newra
//   if it's not contained in the old list.
// also. add new reg_id's to new items if doing reach with edges
//
region_array *
region_array::ra_or_new(region_array ra2, region_array **ranew) {
  int i;
  int this_id; // id # for new regions

  check_sizes_match(ra2,"ra_or_new");
  region_array *newra = new region_array(size, autp);

  for (i=0;i<size;i++) {
    region *regp2 = ra2.regions[i];
    while (regp2 != NULL) {
      region *regp = regions[i];
      Bool contained_in = false;
      while ((regp != NULL) && !contained_in) {
	contained_in = ((*regp2) <= (*regp));
	if (!contained_in) {
	  regp = regp->next;
	}
      }
      if (!contained_in) {
        if (reach_with_edges) {
	  max_reg_id++;
	  this_id = max_reg_id;
	}
	else {
	  this_id = -1;
	}
	region *rppp = regp2->fresh_copy(this_id);
	insertreg(regp2->fresh_copy(this_id), &(newra->regions[i]));
	insertreg(regp2->fresh_copy(this_id), &(regions[i]));

      }
      regp2 = regp2->next;
    }
  } // end for


  *ranew = newra;

  return(this);

}

// return the region_array that is the hull of all states including in 
// original object.
// destroy object in process, and return ptr to itself
region_array *
region_array::ra_hull() {
  int i;
  region *regp, *regp2, *regnext;
  for (i=0;i<size;i++) {
    regp = regions[i];
    if (regp!= NULL) {
      // need to go along list, taking hull with each additional 
      // region for this location
      regp2 = regp->next;
      while (regp2 != NULL) {
	regp->polyp = regp->polyp->hull(regp2->polyp);
	regnext = regp2->next;
	regp2->free(); // global_free(regp2);
	regp2 = regnext;
      }
      regp->next = NULL;
      regions[i]=regp;
    }
  }
  return (this);
}



// N.B. DON'T NEED THIS ANYMORE since doing everything in region arrays
// except for the actual computation for a single region returning a list.
//
// given a  list of regions, make it into a region array wrt the given 
// automaton
// destroy the list in the process!
//
// routine used for reachability, since pre-/post- return lists of regions
region_array *
array_for_regions(region *regp, automaton *automaton_p) {
   region_array *rega = new region_array(automaton_p->nr_locations);
   rega->autp = automaton_p;
   region *rp = regp;
   while (rp != NULL) {
     region *nextp = rp->next;
     insertreg(rp,&(rega->regions[rp->locnp->locn_id]));
     rp = nextp;
   }
   return(rega);
 }


// return region array for the initial regions of aut.
// 
// assumes for now a simple initial specification, i.e. single initial locn
//    and an initial poly.
// 
// in PRE dirn, returns the region array for the final_pred, after 
//     cts evolution in dirn PRE
region_array *
generic_init_region_array(automaton *automaton_p, enum directions dirn,
			  Bool add_cts_evolve) {
  region_array *rega;
  region *tempr;
  if (dirn == POST) {
    rega = new region_array(automaton_p->nr_locations);
    rega->autp = automaton_p;
    insertreg( new region(automaton_p->init_location_p,
		       new polyhedron(*(automaton_p->init_region_p->copy()),true)), 
	      &(rega->regions[automaton_p->init_location_p->locn_id]));
    tempr = rega->regions[automaton_p->init_location_p->locn_id];
    // assume that giving initial states that are final states, but do not
    // lie in their invariants, is BOGUS.
    tempr->restrict_to_invariant();
    if (add_cts_evolve) {
      rega->regions[automaton_p->init_location_p->locn_id] =
	tempr->cts_evolve(dirn);
      tempr->free();
    }
    else {
      rega->regions[automaton_p->init_location_p->locn_id] =
	tempr;
    }
    rega->regions[automaton_p->init_location_p->locn_id]->print();
  }
  else {
    // dirn == PRE
    // so get final array predicate, and do cts_evolve to all regions
    int i;
    region *rp;
    rega = ra_for_pred(automaton_p, final_pred);
    for (i=0;i<automaton_p->nr_locations;i++) {
      // take away old list, incrementally rebuild it with cts-evolved regions
      rp = rega->regions[i];
      rega->regions[i] = NULL;
      while (rp != NULL) {
	rp->restrict_to_invariant();
	if (!rp->is_empty()) {
	  if (add_cts_evolve) {
	    insertreg(rp->cts_evolve(dirn), 
		      &(rega->regions[i]));
	  }
	  else {
	    insertreg(rp->fresh_copy(), 
		      &(rega->regions[i]));
	  }
	  tempr = rp->next;
	}
	rp->free();
	rp = tempr;
      }
    }
  }
  return(rega);
}

// given a region array, prepare it for reachability in direction dirn
// i.e. restrict all regions to the invariant, take cts evolve in correct dirn?
// destroys old object, returns ptr to itself (the new object).
region_array *
region_array::prepare_for_reach(enum directions dirn, int add_cts_evolve) {
  int i;
  region *rp, *tempr;
  for (i=0;i<size;i++) {
    // take away old list, incrementally rebuild it with cts-evolved regions
    rp =regions[i];
    regions[i] = NULL;
    while (rp != NULL) {
      rp->restrict_to_invariant();
      if (!rp->is_empty()) {
	if (add_cts_evolve) {
	  insertreg(rp->cts_evolve(dirn), 
		    &(regions[i]));
	}
	else {
	  insertreg(rp->fresh_copy(), 
		    &(regions[i]));
	}

      }
      tempr = rp->next;
      rp->free();
      rp = tempr;
    }
  }

  return(this);
}

region_array *
init_region_array(automaton *autp, enum directions dirn) {
  // parameter true means "take cts evolve", i.e. really use for init region
  //   not as final region for opposite direction!
   return( generic_init_region_array(autp,dirn,true) );
 }
 
region_array *
final_region_array(automaton *autp, enum directions dirn) {
  // parameter false means "don't take cts evolve", i.e. don't use as
  // initial region, where we have to do first cts evolve, before transitions
  return( generic_init_region_array(autp,opp_dirn(dirn),false) );
}

void
region_array::print() {
  int i;

  for (i=0;i<size;i++) {
    if (regions[i] != NULL) {
      regions[i]->list_print();
    }
  }
}

// restrict each region to its invariant
region_array *
region_array::restrict_to_invariant() {
  int i;
  for (i=0;i<size;i++) {
    if (regions[i]) {
      regions[i]->restrict_to_invariant();
    }
  }
  return(this);
}


//
// perform reachability analysis in the dirn of parameter.
//
// return array of all reachable regions.  
// do not ruin old copy of (initial) array
//
region_array *
region_array::ra_reach(enum directions dirn) {
// at the mo, some memory redundancies, and extra copying.
// the active list is used, then destroyed
  Bool done = false;
  int iter_count = 0;
  region_array *reached = new region_array(this); // accumulates reached states
  reached->autp = autp;
  
  region_array *succ_rega;
  region_array *active;
  region_array new_active_reg = region_array();
  new_active_reg.autp = autp;
  region_array *new_active;

  if (reach_with_edges) {
    max_reg_id = -1;
    did_reach_with_edges = true;
    reached_forward = (dirn == POST);
    reached->ra_assign_reg_ids();
  }
  else {
    did_reach_with_edges = false;
  }
  active  = new region_array(reached);
  new_active = &new_active_reg;

  // second param is how many extra var to allow for. assume worst case is all.
  // this may be redundant.
  make_exist_vectors(var_count,var_count);
    
  while (!active->is_empty()) {
    if ((print_mode == 0) || (print_mode ==2))  {
	cerr << "." << flush;
      }


    if ((print_mode == 1) || (print_mode ==2))  {
	cout << "Iteration: " << iter_count << "\n";
        cout << "Using this newly reached list " << 
	     "=================================\n";
	active->print();
      }

    iter_count++;

    debug_magic_nr = iter_count;


    succ_rega = active->ra_succ(dirn,0);


#ifdef DEBUG_VERIF
    cout << "Got these successors\n"; succ_rega->print();
#endif

    // ra_or_new also assigns reg_ids to the newly active regions    
    reached = reached->ra_or_new(*succ_rega, &new_active);

    active->free();
    succ_rega->free();
    active = new_active;
  }
  cout << "Number of iterations required for reachability: " << iter_count
    << "\n" << flush;
  cerr << "\n" << flush;
  return(reached);
}

//
//
// redo the OR routines so that they call routines for or of region lists
//
// then these called routines can be the basis of a simplification routine
//    that will be used for the elimination of locations
//


// eliminate every variable in the indices list from every region
// can destroy region array
// return ptr to new one (i.e. this)
region_array *
region_array::eliminate_vars(int nr_elim_vars, int *indices) {
  int i;
  region *rp;
  for (i=0;i<size;i++) {
    rp = regions[i];
    while (rp != NULL) {
      if (rp->polyp != NULL) {
	rp->polyp->existent(nr_elim_vars,indices);
      }
      rp = rp->next;
    }
  }
  return(this);
}


region_array *
region_array::eliminate_vars( struct var_name *varnames) {
  int i;
  int elimindices[MAX_VARS];
  int pos = 0;
  struct var_name *x = varnames;

  for (i=0;i<MAX_VARS;i++) {
    elim_var_index[i] = false;
  }
  while (x != NULL) {
    int var_pos = var_index(x->name);
    if (var_pos == -1) {
      cout << "Error: bogus variable name -- " << x->name << "\n";
      exit(1);
    }
    else {
      elim_var_index[var_pos] = true;
    }
    x = x->next;
  }

  for (i=1;i<var_count+1;i++) {
    if (elim_var_index[i]) {
      elimindices[pos] = i-1;
      pos++;
    }
  }
  return ( eliminate_vars(pos,elimindices) );
}

// return ptr to this, after object has all var info eliminated
// i.e. for each location, if any on the list for this locn is nonempty
//    then return ptr to single region with with trivial true matrix
//    and destroy the original list.
// this is much more efficient than calling the eliminate_vars() routine.
region_array *
region_array::eliminate_all_vars() {
  int i;
  matrix rn(1,var_count+2);

  for (i=1;i<var_count+1;i++) rn.p[0][i]=0;
  rn.p[0][0] = rn.p[0][var_count+1] = 1;

  for (i=0;i<size;i++) {
    if (regions[i] != NULL) {
      if (regions[i]->is_empty_list()) {
	// replace with NULL list
	regions[i]->free_list();
	regions[i] = NULL;
      }
      else {
	// replace with list with single R^n element
	// doesn't check the case where the first region's polyp is NULL
	// but this should never happen anymore
	if ((regions[i]->polyp!=NULL) && 
	    (var_count+2 != regions[i]->polyp->constraints()->nbcolumns)) {
	  cerr << "Unexpected dimension of poly during elimination of vars\n"
	    << flush;
	  exit(1);
	}
	else {
	  location *locnptr = regions[i]->locnp;
	  regions[i]->free_list();
	  // assign list the constraint matrix for R^n
	  regions[i] = new region(locnptr,new polyhedron(rn,true));
	}
      }
    }
  }	
  return this;
}

region_array *
region_array::eliminate_non_params() {
  // need to check indices, vs var numbers ...
  int *indices = (int *) malloc (var_count * sizeof(int));
  int i;
  int nr_non_params = 0;
  
  for (i=0;i<var_count;i++) {
    if (var_type(i+1) != V_PARAMETER) {
      // special test for timing case!!
      indices[nr_non_params] = i;

      nr_non_params++;
    }
  }
  return(eliminate_vars(nr_non_params,indices));
}

// set all names to some constant ("")
// dangerous. makes automaton unusable, since names no longer exist.
void
region_array::make_names_constant() {
  int i;
  region *rp;
  for (i=0;i<size;i++) {
    rp = regions[i];
    // this will set the names for all regions using this location.
    if (rp != NULL) {
      global_free_str(rp->locnp->name);
      rp->locnp->name = str_preserve("");
    }
  }
}



// operate on object. return ptr to itself.
// eliminate locations for the process proc_nr (start numbering from 0)
region_array *
region_array::eliminate_location_for_proc(int proc_nr) {
  int i;
  region *rp;
  for (i=0;i<size;i++) {
    rp = regions[i];
    while (rp != NULL) {
      rp->elim_loc_name_for_proc(proc_nr);
      rp = rp->next;
    }
    rp  = (region *) NULL;
    region_or(&rp,regions[i]);
    // memory leak here for regions[i]?
    // plugged 4/17
    if (regions[i] != NULL) {
      regions[i]->free_list();
    }
    regions[i] = rp;
  }
  return(this);
}


// temporary routine
// for now, just eliminate all locations.
// return region list for result
// can destroy the object in process
region *
region_array::eliminate_locations() {
  // use the OR operation on lists.
  region *rp = NULL;
  int i;
  make_names_constant();
  for (i=0;i<size;i++) {
    region_or(&rp,regions[i]);
  }
  return(rp);

}

// temporary routine
// for now, just eliminate all locations.
// return region list for result
// can destroy the object in process
region *
region_array::merge_locations_by_name() {
  // use the OR operation on lists.
  region *rp = NULL;
  int i;

  for (i=0;i<size;i++) {
    region_or(&rp,regions[i]);
  }
  return(rp);

}



// temporary routine
// for now, just eliminate all locations with str, and return list of them
// return region list for result
// can destroy the object in process
//
// BEWARE: also destroys the automaton's location names.  NASTY.
//   FIX THIS.
region *
region_array::eliminate_locations_with_str(char *str) {
  // use the OR operation on lists.
  region *rp = NULL;
  int i;
  for (i=0;i<size;i++) {
    if (regions[i] != NULL) {
      // test whether input str in the name anywhere (very loose requirement)
      if (strstr(regions[i]->locnp->name,str)) {
	regions[i]->locnp->name = str_preserve("");
	region_or(&rp,regions[i]);
      }
    }
  }
  return(rp);
}

// simplify each list of regions
void
region_array::simplify() {
  int i;
  for (i=0;i<size;i++) {
    if (regions[i] != NULL) {
      // mem leak. plugged 12/5/96
      region *temp = regions[i];
      regions[i] = regions[i]->simplify();
      if (temp) {
	temp->free();
      }
    }
  }
}


region *
region_array::eliminate_all() {
  eliminate_non_params();
  simplify();
  return(eliminate_locations());
}



// elim non-param and loc. according to global controlling params
// massive memory leaks at the moment
// prints the output
// and returns ptr to new list of regions for simplified output
region *
region_array::output_elim() {
  int i;
  region_array *rap;
  region *rp;
  if (elim_vars_by_list) {
    cout << "Eliminating variables by user-specified list\n";
    int elimindices[MAX_VARS];
    int pos = 0;
    for (i=1;i<var_count+1;i++) {
      if (elim_var_index[i]) {
	elimindices[pos] = i-1;
	pos++;
      }
    }
    rap = eliminate_vars(pos,elimindices);
  }
  else {
    switch(elim_all_vars) {
    case 1:
      rap = eliminate_non_params();
      break;
    case 2:
      rap = eliminate_all_vars();
    case 0: default: // don't eliminate anything
      rap = this;
      break;
    }
  }
  rap->print();
  // mem leak.
  rap->simplify();

  // now do the elim by locations
  if (elim_locs_by_list) {
    cout << "Eliminating locations by user-specified list of processes\n";
    for (i=0;i<MAX_PROCS;i++) {
      if (elim_proc_index[i]) {
	cout << "Eliminating locations for proc " << i << "\n";
	rap = rap->eliminate_location_for_proc(i);
      }
    }
    rap->simplify();
    rap->print();
    region *rp = rap->merge_locations_by_name();
    cout << "=======================================\n";
    cout << "Final simplified output\n" << flush;
    rp->print();
    return(rp);
  }
  else {
    // use the elim all or none option
    if (elim_all_locns) {
      rp = rap->eliminate_locations_with_str("");
      cout << "Here's the simplified string\n" << flush;
      if (rp==NULL) {
	cout << "System verified\n" << flush;
      }
      while (rp != NULL) {
	rp->print();
	rp = rp->next;
      }
      cout << "End of simplification\n";
      return(rp);
    }
    else {
      rap->print();
      return(NULL);
    }
  }
} 


// THIS routine used by print_simplified_stmts, i.e. print omit ...

// elim location info only. according to global controlling params
// elim_proc_index and elim_locs_by_list and elim_all_locns
// massive memory leaks at the moment
// does no printing
// and returns ptr to new list of regions for simplified output
// WARNING destroys automaton location names
// MUST call copy_location_names() before this procedure, then
// call rsetore_location_names() after it.
void
region_array::elim_locs_and_print() {
  int i;
  region_array *rap;
  region *rp;

  rap = new region_array(this);
  if (elim_locs_by_list) {
    for (i=0;i<MAX_PROCS;i++) {
      if (elim_proc_index[i]) {
	rap = rap->eliminate_location_for_proc(i);
      }
    }
    rap->simplify();
    region *rp = rap->merge_locations_by_name();
    if (rp != NULL) {
      rp->min_print();
      rp->free();
    }
    else {
      cout << "     False\n";
    }
    // global_free(rp);
  }
  else {
    // use the elim all or none option
    if (elim_all_locns) {
      rp = rap->eliminate_locations_with_str("");
      if (rp != NULL) {
        rp->min_print(); 	
        rp->free(); 
      }
      else {
	cout << "     False\n";
      }
      // global_free(rp);
    }
    else {
      rap->print();
    }
  }
  rap->free();
} 


// destroys the current object, places result of eliminating all
// locations in every location.
void
region_array::elim_locs_and_store(automaton *aut) {
  int i;
  region_array *rap;
  region *rp;

  rap = new region_array(this);
  if (elim_locs_by_list) { 
    cout << "OMIT expression not implemented for lists\n";
    exit(1);
    for (i=0;i<MAX_PROCS;i++) {
      if (elim_proc_index[i]) {
	rap = rap->eliminate_location_for_proc(i);
      }
    }
    rap->simplify();
    region *rp = rap->merge_locations_by_name();
    if (rp != NULL) {
      rp->min_print();
      rp->free();
    }
    else {
      cout << "     False\n";
    }
    // global_free(rp);
  }
  else {
    // use the elim all or none option
    if (elim_all_locns) {
      rp = rap->eliminate_locations_with_str("");
      if (rp != NULL) {
        rp->minimize();
      }
      else {
	rp = new region(&(aut->locations[0]),empty_polyhedron(var_count+2));
      }
      for (i=0;i<rap->size;i++) {
	if (rap->regions[i] != NULL) {
          rap->regions[i]->free();
          // global_free(rap->regions[i]);
        }
	if (regions[i] != NULL) {
	  regions[i]->free();
	  // global_free(regions[i]);
	}
	rap->regions[i] = rp->fresh_copy_list();        
	region *rp2 = rap->regions[i];
	while (rp2 != NULL) {
	  rp2->locnp = &(aut->locations[i]);
	  rp2 = rp2->next;
       }
	regions[i] = rap->regions[i];
      }
      // global_free(rp);
    }
  }


} 



// exact containment check
Bool 
region_array::ra_exact_is_in(region_array *rap) {
  int i = 0; 

  Bool in_so_far = true;
  while (in_so_far && (i < size)) {
//  cout << "checking i = " << i << endl<<flush;
    if (regions[i] != NULL) {

      in_so_far = regions[i]->exact_is_in(rap->regions[i]);
    }
    i++;
  }
  return(in_so_far);
}
  
// exact equivalence check
Bool 
region_array::ra_exact_is_equal(region_array *rap) {
  int i = 0; 

  Bool eq_so_far = true;
  while (eq_so_far && (i < size)) {
    eq_so_far = regions[i]->exact_is_equal(rap->regions[i]);
    i++;
  }
  return(eq_so_far);
}

// exact strict < test (object's states strictly less than rap's states)
Bool
region_array::ra_exact_is_lt( region_array *rap) {
 return( ra_exact_is_in(rap) && !rap->ra_exact_is_in(this) );
}

// exact strict > test (object's states strictly greater than rap's states)
Bool
region_array::ra_exact_is_gt( region_array *rap) {
 return( rap->ra_exact_is_lt(this) );
}

void
region_array::print_size() {
  int i;
  int nr_non_empty = 0;     // nr locn for which region list is non empty
  int tot_size = 0;         // nr regions altogether in region_array
  for (i=0;i<size;i++) {
    if (regions[i] != NULL) {
      nr_non_empty++;
      tot_size += regions[i]->list_size();
    }
  }
  if (tot_size == 0) {
    cout << "Empty region array\n" << flush;
  }
  else {
    cout << "Size of region array:\n\t" << nr_non_empty 
      << " locations (out of " << size << " total) have states\n\t" << tot_size << " convex predicates altogether\n";
   cout << "\tAverage convex predicates per location with states = " << tot_size/nr_non_empty;
     cout << "\n\tAverage over all locations = " << tot_size/size 
      << "\n" << flush;
  }
}

void
region_array::ra_assign_reg_ids() {
  int i;
  for (i=0;i<size;i++) {
    if (regions[i] != NULL) {
      regions[i]->assign_reg_ids();
    }
  }
}

// to generate pts, do the following:
// 1) add region * pt to parameter list of print_trace_from
// 2) add routine to select a pt in a region
//       tricky part is dealing with strict inequalities
// 3) find predecessor pts before recursively calling next print_trace_from()
//        find predecessor in the region with ID this_reg->pred_reg_id
//        by doing first time pred, then pred. across the particular transition
//        then pick a pt there.  now do instantaneous transition, and intersect
//        with time pred from above, choose a pt.  print both newly found pts.
//        later also figure out the time elapsed between the pts.
// Most routines written already, e.g. single_event_succ(i,dirn), 
//        cts_evolve(dirn)
// Need to write routines for 
// 1) picking pt in a region
// 2) finding intermediate pt and pred pt (by calling other routines).
// Problem: matching the PRE's for a transition  with its corresponding POST
//      out of another location.
// Use hack for now. take all POST's until find one that intersects the 
//      timed_pred region.  Take PRE of it, and see if it intersects.  It 
//      eventually will.  Choose a pt in the intersection.

region *
selected_pt_in(region *set_region) {
  // yields a region for a pt in set_region
  // assume the set_region is non-empty
  // find a pt, i.e. row in frame for which first element is 1, last is nonzero
  // and for which the eps_slack variable is non-zero.
  matrix *fr = set_region->polyp->nplus1_form->frame();
  int nrows = fr->nbrows;
  int ncols = fr->nbcolumns;
  int epos = set_region->polyp->eps_pos;
  set_region->polyp->fix_strict();
  Bool strict_region = set_region->polyp->strict;
  int i;
  for (i=0;i<nrows-1;i++) {
    if ((fr->p[i][0] == 1) && (fr->p[i][ncols-1] != 0) &&
	(!strict_region || (fr->p[i][epos] != 0))) 
      break;
  }
  // now i is pos of row for a pt
  int ptrow = i;

  // need one constraint per variable
  int nr_cons = ncols-3;
  matrix *cons = new matrix(nr_cons+1,ncols);
  // make the cons matrix
  cons->blank();
  cons->set_last_row();
  int cons_row = 0;
  for (int j=1;j<ncols-1;j++) 
    if (j!= epos) {
      cons->p[cons_row][j] = fr->p[ptrow][ncols-1];
      cons->p[cons_row][ncols-1] = - fr->p[ptrow][j];
      cons_row++;
    }

  // constructor for poly assumes in nplus1_form
  polyhedron * pt_cons = new polyhedron(*cons,true);
  return( new region(set_region->locnp, pt_cons) );
}


// given a pred_region, and int_pred_set (i.e. set of points both in image
//   of instantaneous (assumed POST) trans, and time-pre of a pt.
// find a transition # in direction dir (assumed PRE) from int_pred_set to 
//     pred_region
// used in find_pts();
int
pre_tr_nr_for(region *pred_reg, region *int_pred_set, enum directions dir) {
  // algo. try each transition into pred_reg, until pre(int_pred_set)
  //    for a  single trans intersects pred_reg
  int i;

  int nr_tr = (dir == PRE) 
    ? int_pred_set->locnp->nr_pre_trans
      : int_pred_set->locnp->nr_post_trans;
  transition *tr_array = (dir == PRE)
    ? int_pred_set->locnp->pre_trans
      : int_pred_set->locnp->post_trans;
  for (i=0;i<nr_tr;i++) {
    region * pre_set = int_pred_set->single_event_succ(i,dir);
    region * int_reg = pre_set->intersect(pred_reg);
    if (!int_reg->is_empty()) {
      pre_set->free();
      int_reg->free();
      return(i);
    }
    pre_set->free();
    int_reg->free();
  }
  // shouldn't ever reach here.
  cerr << "Internal error: Bad call of pre_tr_nr_of()\n" << flush;
  exit(1);
}


// given 
//  - pt,  defined by constraints on all variables except eps_slack
//  - slopes, 
// returns a possible time taken in the slopes direction to get from 
//   the origin to pt.
// denom_lcm, indicates how much time has to be scaled up.
float
time_evolving(lib_polyhedron *pt, lib_polyhedron *slopes, 
	      int slope_eps_pos) {
  int i;


  lib_polyhedron *int_slope  = slopes->intersection(*(pt->constraints()));
  matrix *int_slope_fr = int_slope->frame();


  //  for (i=0;i<int_slope_fr->nbrows-1;i++) {
  // reaching the final row of 10001 is OK, if it's the only row, it means take
  // the origin
  for (i=0;i<int_slope_fr->nbrows;i++) {
    // choose a vertex
    if ((int_slope_fr->p[i][0]==1 )
	 && (int_slope_fr->p[i][int_slope_fr->nbcolumns-1] !=0)) {
      // cout << "choosing vertex in row i = " << i << endl << flush;
      // cout << "... is ... " << float(int_slope_fr->p[i][slope_eps_pos] ) / 
      //	int_slope_fr->p[i][var_count+1] << endl << flush;

      return( float(int_slope_fr->p[i][slope_eps_pos] ) / 
	       int_slope_fr->p[i][var_count+1]);
    }
  }
  cerr << "Interal Error in trace inter-transition time computation\n";
  exit(1);
}

int
vertex_row(matrix *mat) {
  int i;
  for (i=0;i<mat->nbrows-1;i++) {
    if ((mat->p[i][0] ==1) && (mat->p[i][mat->nbcolumns-1] !=0))
      return i;
  }
  cout << "Error. No rows in frame are vertexes\n";
  exit (1);
}


// return difference between pt2 and pt1
// assumed to be points specifying coordinates for every variable
// the frame representation should be a single vertex!
lib_polyhedron *
diff_bw(polyhedron *pt1, polyhedron * pt2) {
  int j;
  int pt1_row, pt2_row;
  lib_polyhedron *res;
  matrix * vert_diff = new matrix(2,var_count+2);
  vert_diff->blank();
  vert_diff->set_last_row();
  pt1->minimize();
  pt2->minimize();
  matrix * pt1_fr = pt1->nplus1_form->frame();
  matrix * pt2_fr = pt2->nplus1_form->frame();
  pt1_row = vertex_row(pt1_fr);
  pt2_row = vertex_row(pt2_fr);
  coeff den1 = pt1_fr->p[pt1_row][var_count+1];
  coeff den2 = pt2_fr->p[pt2_row][var_count+1];


  for (j=1;j<=var_count;j++) {
    vert_diff->p[0][j] = myadd(mymult(pt2_fr->p[pt2_row][j],den1),
			       - mymult(pt1_fr->p[pt1_row][j],den2));
  }
  // to indicate it is a point in the frame
  vert_diff->p[0][0] = 1;

  vert_diff->p[0][pt1->eps_pos] = 0;
  vert_diff->p[0][var_count+1] = mymult(den1,den2);


  // construct result using the frame matrix representation
  res = new lib_polyhedron(*vert_diff,false) ;

  res = res->existent(pt1->eps_pos,0);

  return res;
}






// return a poly for the slopes, where the eps coordinate indicates
// the elapsed time.
// Algo:
// take the slope matrix, intersect with eps = 1.
// take its frame.
// Use elements of frame as generators, making vertices into rays, by
// multiplying all other vertices by the scalar factor in its denom.
lib_polyhedron *
slope_poly(location * locnp, enum directions dirn) {
  int i,j;

  int this_eps_pos = locnp->post_slopes_matrix->eps_pos;
  matrix eps_one(2,var_count+2);
  eps_one.blank();
  eps_one.set_last_row();
  eps_one.p[0][this_eps_pos] = 1;
  eps_one.p[0][var_count+1] = -1;

  lib_polyhedron *new_slopes = 
    (dirn == POST)  ? locnp->post_slopes_matrix->nplus1_form->copy()
      : locnp->pre_slopes_matrix->nplus1_form->copy();
  new_slopes = new_slopes->existent(this_eps_pos,0);


  new_slopes = new_slopes->existent(this_eps_pos,0);
  new_slopes = new_slopes->intersection(eps_one);


  // compute lcm of vertex denominators
  matrix * fr = new_slopes->frame();

  coeff denom = 1;
  for (i=0;i<fr->nbrows-1;i++) {
    if ( (fr->p[i][0]==1) && 
	 (fr->p[i][var_count+1]!=0)) {
      // a vertex
      denom = lcm_coeff(denom,fr->p[i][var_count+1]);
    }
  }

  // convert vertices into rays
  for (i=0;i<fr->nbrows-1;i++) {
    if ( (fr->p[i][0]==1) && 
	 (fr->p[i][var_count+1]!=0)) {
      coeff scalar = denom / fr->p[i][var_count+1];
      for (j=1;j<var_count+1;j++) {
	fr->p[i][j] = mymult(fr->p[i][j],scalar);
      }
      fr->p[i][this_eps_pos] = denom;
      fr->p[i][var_count+1] = 0;
    }
  }
  matrix slope_fr(fr->nbrows+1,fr->nbcolumns);
  slope_fr.blank();
  slope_fr.set_last_row();
  // copy all rows, including the last, effectively adding the origin as vertex
  for (i=0;i<fr->nbrows;i++) {
    for (j=0;j<fr->nbcolumns;j++) {
      slope_fr.p[i][j] = fr->p[i][j];
    }
  }
  
   lib_polyhedron *res = new lib_polyhedron(slope_fr,false);

  return( res);
}


float
time_diff(region *pt1, region *pt2, enum directions dirn) {

  lib_polyhedron *slopes = slope_poly(pt1->locnp,dirn);
  lib_polyhedron * diff_vec = diff_bw(pt1->polyp,pt2->polyp);
  return( time_evolving(diff_vec,slopes, pt1->polyp->eps_pos) );
}











// given
// - pred_region, region to find the next significant pt in
// - pred_trans_in transition from pred_region to the pt
// - pt to backtrack from 
// find
// - timed_pred_pt, point that is backtracked via time step from pt
// - pred_pt, predecessor pt via transition pred_trans_in from timed_pred_pt
void
find_pts(region *pred_region, transition * pred_trans_in, 
	 region *pt, region **time_pred_pt_p, region **pred_pt_p) {
  // reached_forward is global indicating whether last reachability
  //    graph was created using forward reachability

  region *time_pred_set;
  enum directions  search_dir = reached_forward ? PRE : POST;


  time_pred_set = pt->cts_evolve(search_dir);


  int i; 
  int nr_trs, tr_nr;
  transition *tr_array;
  // search_dir is direction we now have to search, NOT the direction used
  // in the reachability.
  if (search_dir == PRE) {
    nr_trs =  pred_region->locnp->nr_post_trans;
    tr_array = pred_region->locnp->post_trans;
  }
  else {
    nr_trs = pred_region->locnp->nr_pre_trans;
    tr_array = pred_region->locnp->pre_trans;
  }
  for (i=0;i<nr_trs;i++) 
    if (&(tr_array[i])==pred_trans_in) {
      tr_nr = i;
      break;
    }
  if (i== nr_trs) {
    cout << "Didn't find the trans\n" << flush;
  }
  region *post_pred_set = pred_region->single_event_succ(tr_nr,opp_dirn(search_dir));

  region *intermed_pred_set = post_pred_set->intersect(time_pred_set);
  post_pred_set->free();
  *time_pred_pt_p = selected_pt_in(intermed_pred_set);
  int pre_tr_nr = pre_tr_nr_for(pred_region,intermed_pred_set, search_dir);
  region *pre_pred_set = 
    (*time_pred_pt_p)->single_event_succ(pre_tr_nr, search_dir);
  region *trace_pre_pred_set = pre_pred_set->intersect(pred_region);
  pre_pred_set->free();
  *pred_pt_p = selected_pt_in(trace_pre_pred_set);


  float res_diff = time_diff(*time_pred_pt_p,pt,POST);

  intermed_pred_set->free();
  trace_pre_pred_set->free();
  time_pred_set->free();
}


void
foutput(float ft) {
  if (ft >= 1.0e+06 | (ft > 0 & ft < 1e-03)) {
    cout.setf(ios::scientific, ios::floatfield);
    cout << ft << "\n";
    cout.setf(0,ios::floatfield);
  }
  else {
    int decplaces = 3;
    int multfactor = 1000;
    double integral, fraction;
    fraction = modf(ft,&integral);
    fraction *= multfactor;
    cout.fill('0');
    cout.setf(ios::fixed);
    cout << setprecision(0) <<  integral << "." ;
    cout.setf(ios::floatfield);
    if (fraction >= 100) {
      // i.e., had three digits
      cout << setw(3) << setprecision(3) << fraction ;
    }
    else {
      if (fraction >= 10) {
         // i.e. has two digits
         cout << "0" << setw(2) << setprecision(2) << fraction ;
      }
      else {
         // has one digit
         cout << "00" << setw(1) << setprecision(1) << fraction;
      }
    }
    cout.setf(0,ios::floatfield);
    cout.precision(6);
    cout.fill(' ');
  }
}



// print trace, using reach info in object, ptrs to regs in id_to_reg, 
// flag print_forward indicates whether to print forward or backward trace
// wrt the search direction.
// i.e. print_forward for a backwards search prints an absolute backwards trace
void
region_array::print_trace_from(region **id_to_reg, Bool print_reverse,
			       int target_id, region *pt, 
			       float *cum_time) {
  region *this_reg = id_to_reg[target_id];
    region * time_pred_pt = NULL, *pred_pt = NULL;



  float time_elapsed;
  enum directions time_dirn = reached_forward ? POST : PRE;
  if (this_reg->pred_reg_id != -1) {
    // has a real predecessor
    find_pts(id_to_reg[this_reg->pred_reg_id], 
	     this_reg->trans_in,
	     pt, 
	     &time_pred_pt, &pred_pt);
    time_elapsed = time_diff(time_pred_pt, pt,time_dirn);	
  }
  else {
    // the first point, so could have a time predecessor
    // generated before call of reach()
    if (pt->is_contained_in(*(init_regn->regions[pt->locnp->locn_id]))) {
      time_pred_pt =pt->fresh_copy();
    }
    else {
      // find a predecessor pt within init_regn
      region *time_pred_set = pt->cts_evolve(opp_dirn(time_dirn));
      region *init_time_pred_set = time_pred_set->intersect(init_regn->regions[time_pred_set->locnp->locn_id]);
      time_pred_pt = selected_pt_in(init_time_pred_set);
      time_pred_set->free();
      init_time_pred_set->free();
      time_elapsed = time_diff(time_pred_pt, pt,time_dirn);	
    }
  }

  if (!print_reverse) {
    if (reached_forward) {
      if (this_reg->pred_reg_id != -1) {
	// has a real predecessor
	if (!time_pred_pt->is_equal(*pt)) {
	  cout << "---------------\n";
	  cout << " VIA ";
	  foutput(time_elapsed); cout << " time units \n";
	  cout << "---------------\n";
	  cout << "Time: ";   foutput(*cum_time);    cout << "\n";
	  time_pred_pt->min_print();
	}
	time_pred_pt->free();
      }
      if (this_reg->trans_in != NULL)
	this_reg->trans_in->label_print();
    }

    cout << "Time: ";   foutput(*cum_time);    cout << "\n";
    pt->min_print();

    if (!reached_forward) {
      if (this_reg->pred_reg_id != -1) {
	// has a real predecessor

	if (!time_pred_pt->is_equal(*pt)) {
	  *cum_time += time_elapsed;
	  cout << "---------------\n";
	  cout << " VIA "; foutput(time_elapsed); 
	  cout << " time units \n";	
	  cout << "---------------\n";
	  cout << "Time: "; foutput(*cum_time); cout  << "\n";
	  time_pred_pt->min_print();
	  time_pred_pt->free();

	}
	this_reg->trans_in->label_print();
      }
    }
  }
  
  if (this_reg->pred_reg_id != -1) {
    print_trace_from(id_to_reg, print_reverse, this_reg->pred_reg_id,pred_pt, cum_time);
  }


  if (print_reverse) {
    if (reached_forward) {
      if (this_reg->pred_reg_id != -1)  {
	this_reg->trans_in->label_print();
      }
      if (!time_pred_pt->is_equal(*pt)) {
	cout << "Time: ";  foutput(*cum_time); cout  << "\n";
	time_pred_pt->min_print();
	cout << "---------------\n";
	cout << " VIA "; foutput(time_elapsed); cout << " time units \n";	
	time_pred_pt->free();
	cout << "---------------\n";
	*cum_time += time_elapsed;
      }
    }
    cout << "Time: "; foutput(*cum_time); cout  << "\n";
    pt->min_print();

    if (!reached_forward) {
      if (this_reg->pred_reg_id != -1)  {
	if (!time_pred_pt->is_equal(*pt)) {
	  cout << "---------------\n";
	  cout << " VIA ";
	  *cum_time += time_elapsed;
	  foutput(time_elapsed);
	  cout << " time units \n";
	  cout << "---------------\n";
	  time_pred_pt->min_print();
	}
	time_pred_pt->free();
	this_reg->trans_in->label_print();
      }
    }
  }
}

void
region_array::print_trace(region **id_to_reg, region_array *int_ra,
		    Bool print_reverse) {
  int i = 0;
  Bool found = false;
  region *target_reg;
  region *regp;
  int min_so_far;
  region *init_pt;
  float cum_time = 0;

  // select a candidate final region
  // use minimal reg_id as criterion for selection
  min_so_far = max_reg_id + 2;
  for (i=0;i<int_ra->size;i++) {
    regp = int_ra->regions[i];
    while (regp != NULL) {
      if (!regp->is_empty()) {
	// see if less than previous minimum
	if (regp->reg_id < min_so_far) {
	  target_reg = regp;
	  min_so_far = regp->reg_id;
	  found = true;
	}
      }
      regp = regp->next;
    }
  } 

  if (!found) {
    cerr << "Called print_trace when no traces exist\n" << flush;
  }
  else {
    // found a region to search backwards from, namely target_reg

    init_pt = selected_pt_in(target_reg);
/*
   // since final reg is explicitly intersected with the reachable
   // region, we can be sure that the trace will take into account an
   // final time event required in the reachability analysis
   // so the trace will include it.

    if (!print_reverse) {
      // maybe have to print something for the initial time event
      // since not printing in reverse, first event is in
      // the 'final' region of the reachability analysis
      region_array *init_backtrack_regn = 
      region_array init_pt_ra(size,  init_pt);
      init_pt_ra.autp = autp;
      if (!init_pt_ra.ra_in(init_regn)) {
	enum directions  time_dirn = reached_forward
						    ? POST
						    : PRE ;
	region *time_pred_set = init_pt->cts_evolve(time_dirn);	
	region *init_time_pred_set = 
	  time_pred_set->intersect(init_regn->
				   regions[time_pred_set->
					   locnp->locn_id]);
	region *time_pred_pt = selected_pt_in(init_time_pred_set);
	time_pred_set->free();
	init_time_pred_set->free();
	cout << "Time: ";
	foutput(0);
	cout << "\n";
	time_pred_pt->print();
	float time_elapsed = time_diff(time_pred_pt, init_pt,time_dirn);	
	cout << "  VIA "; foutput(time_elapsed); cout << " time units\n";
	cum_time = time_elapsed;
	}
      }
*/
   print_trace_from(id_to_reg, print_reverse, target_reg->reg_id,  init_pt,&cum_time);
    cout << " ============ End of trace generation ============\n";
  }

}



// generate an error trace, given
// 1) object regions array for the reached region, with reg_id's, etc
// 2) final_ra, the regions array for the final regions
// N.B. in object ra, pred_reg_id == -1 means initial region
//  N.B. assumes the initial region has had time added to it already.
//   hence, need to pass as argument the real initial region, and
//   add time-step to get a state in the 'initial' region used for 
//   reachability.
void 
region_array::generate_trace_to(region_array *final_ra) {
  region_array *int_ra;  // the intersection of object with final_ra
  Bool print_reverse;    // whether to print in reverse order
                         // of when states are generated

  // assign the global flag so all regions routines know
  generating_trace = true;



  cout << " ====== Generating trace to specified target region ========\n";
  if (!did_reach_with_edges) {
    cout  << "Cannot generate traces unless "
      << "reachability performed with trace information\n" << flush;
    cerr  << "Cannot generate traces unless "
      << "reachability performed with trace information\n" << flush;
  }
  else {
    // ra_and assigns the id's of the object's regions to the intersected parts
    int_ra = ra_and(*final_ra);
    if (int_ra->ra_empty()) {
      cout << "No path to indicated target region\n";
    }
    else {
      // generate the trace
      // 1) create table pting from reg_id to the region itself
      // 2) recursively print edges
      int i;
      region *regp;
      region **id_to_reg = 
	(region **) malloc( sizeof(region *) * (max_reg_id+1));
      // 1) create table in id_to_reg
      for (i=0;i<size;i++) {
	regp = regions[i];
	while (regp != NULL) {
	  if ((regp->reg_id <0) || (regp->reg_id > max_reg_id)) {
	    cerr << "Reg id out of range.  Aborting trace generation.\n" 
	      << flush;
	  }
	  id_to_reg[regp->reg_id] = regp;
	  regp = regp->next;
	}
      }


      // 2) generate trace recursively
      print_reverse = (   (reached_forward && generate_forward_trace) 
	  || (!reached_forward && !generate_forward_trace));
      
      print_trace(id_to_reg,int_ra,print_reverse);
      global_free(id_to_reg);
    }
    int_ra->free();
  }
  // done with trace generation, clear flag
  generating_trace = false;
}


